/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2003 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */

#include "Mutex.hpp"
#include "Logsystem.hpp" 

Guard::Guard(pthread_mutex_t* m)
{
  if (m==NULL) {
    Logsystem::debug_logger()->Tstamp() 
      << "    Guard::Guard: NULL mutex\n";
    return;
  }
  _mutex=m;
  if (pthread_mutex_lock(_mutex)!=0)
    Logsystem::debug_logger()->Tstamp() 
       << "    Guard::Guard:: unable to get lock\n";
  return;
}

Guard::~Guard()
{
  if (_mutex==NULL) {
    Logsystem::debug_logger()->Tstamp() 
      << "    Guard::~Guard: NULL mutex\n";
    return;
  }
  if (pthread_mutex_unlock(_mutex)!=0)
    Logsystem::debug_logger()->Tstamp() 
      << "    Guard::Guard:: unable to get lock\n";
  return;
}

Mutex::Mutex()
{
  _mutex=new pthread_mutex_t;
  pthread_mutex_init(_mutex, NULL);

  _cond_ = new pthread_cond_t;
  pthread_cond_init(_cond_,NULL);
}

Mutex::~Mutex()
{
  pthread_mutex_destroy(_mutex);
  delete _mutex;
  pthread_cond_destroy(_cond_);
  delete _cond_;
}

int Mutex::lock()
{
  if (pthread_mutex_lock(_mutex)!=0){
    Logsystem::debug_logger()->Tstamp() 
      << "    Mutex::lock: unable to get lock\n";
    return -1;
  }else
    return 0;
}

int Mutex::trylock()
{
  int ret = pthread_mutex_trylock(_mutex);
  switch(ret){
  case EINVAL:
    Logsystem::debug_logger()->Tstamp() 
      << "Error in Mutex::trylock():The value does not refer to an initialized mutex object.\n";
    return -1; break;
  case EBUSY:
    Logsystem::debug_logger()->Tstamp() 
      << "Error in Mutex::trylock(): Already locked\n";
    return -1;
  case EAGAIN:
    cerr << "Error in Mutex::trylock():maximum number of recursive locks has been  exceeded.";
    return -1; break;
  case 0: 
    return 0; break;
  default:
    Logsystem::debug_logger()->Tstamp() 
      << "Error in Mutex::trylock: unable to get lock by undef error.\n";
    return ret;
  }

}

int Mutex::unlock()
{
  if (pthread_mutex_unlock(_mutex)!=0){
    cerr << "    Mutex::unlock: unable to get lock\n";
    return -1;
  }else
    return 0;
}

/** Conditional wait functions---
 * Inserted by M.Sekido 
 **---------------------------*/
const int Mutex::cond_wait(){
  //  if(trylock() <0) cerr << "Error trylocl() in Mutex::cond_timewait()\n";
  if(pthread_cond_wait(_cond_, _mutex) !=0){
    cerr << " Error in Mutex::cond_wait(): failed cond_wait()\n";
    //    unlock();
    return -1;
  }
  //  unlock();
  return 0;
}

const int Mutex::cond_timewait(const struct timespec *abstime){
  //  if(trylock() <0) cerr << "Error trylocl() in Mutex::cond_timewait()\n";
  int ret = pthread_cond_timedwait(_cond_, _mutex,abstime);
  switch(ret){
  case ETIMEDOUT:
    Logsystem::debug_logger()->Tstamp() 
      << "Info: Mutex::cond_timewait(): Condition was not send by "
      << asctime(gmtime(&(abstime->tv_sec))) << "\n";
    //    unlock();
    return -1;
    break;
  case EINTR:
    Logsystem::debug_logger()->Tstamp() 
      << "Info: Mutex::cond_timewait(): Interupted by signal.\n";
    //    unlock();
    return -1;
    break;
  case 0:
    //    unlock();
    return 0;
  default:
    Logsystem::debug_logger()->Tstamp() 
      << "Error in Mutex::cond_wait(): failed cond_timedwait() by some reason.\n";
    //    unlock();
    return -1;
    break;
  }
   
}

/** Simplify the argument for cond_timewait_ */
const int Mutex::cond_timewait_2(const int32_t msec){
  if(msec < 0){
    Logsystem::debug_logger()->Tstamp() 
      << "Error at Mutex::cond_timewait_2(): msec(" << msec 
      << ") is negative, unexpectedly\n";
    return -1;
  }
  struct timeval now;
  struct timespec timeout;
  time_t  isec = msec / 1000;
  int32_t usec = msec - isec * 1000;
  usec *= 1000;
  gettimeofday(&now,0);
  timeout.tv_sec = now.tv_sec + isec;
  timeout.tv_nsec = (now.tv_usec + usec) * 1000;
  int ret = pthread_cond_timedwait(_cond_, _mutex, &timeout);
  switch(ret){
  case ETIMEDOUT:
    Logsystem::debug_logger()->Tstamp() 
      << "Error in Mutex::cond_timewait(): Condition was not send by "
      << asctime(gmtime(&(timeout.tv_sec))) << "\n";
    //    unlock();
    return -1;
    break;
  case EINTR:
    Logsystem::debug_logger()->Tstamp() 
      << "Error in Mutex::cond_timewait(): Interupted by signal.\n";
    //    unlock();
    return -1;
    break;
  case 0:
    //    unlock();
    return 0;
  default:
    Logsystem::debug_logger()->Tstamp() 
      << "Error in Mutex::cond_wait(): failed cond_timedwait() by some reason.\n";
    //    unlock();
    return -1;
    break;
  }
   
}

const int Mutex::cond_signal(){
  if(pthread_cond_signal(_cond_) !=0){
    Logsystem::debug_logger()->Tstamp() 
      << " Error in Mutex::cond_signal(): failed cond_signal()\n";
    return -1;
  }else
    return 0;
}

const int Mutex::cond_broadcast(){
  if(pthread_cond_broadcast(_cond_) !=0){
    Logsystem::debug_logger()->Tstamp() 
      << " Error in Mutex::cond_broadcast(): failed cond_broadcast()\n";
    return -1;
  }else
    return 0;
}
