/*
 *  File.h
 *
 *  Created by David Lapsley on Fri Feb 20 2004.
 *
 */

/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2003 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */

/** FileSocket class.
 * This class encapsulates a C file. 
 * @author David Lapsley (dlapsley@haystack.mit.edu)
 */

#ifndef FILE_SOCKET_H
#define FILE_SOCKET_H

#include <Socket.h>
#include <common.h>
#include <string>
#include <iostream>
#ifdef LINUX
#include <unistd.h>
#endif // LINUX
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <Object.h>

using namespace std;

/** FileSocket class.
 * This class encapsulates a C file. It inherits from Socket and implements
 * the same interface as all other Socket classes. Note that, unlike
 * other XSocket classes, the FileSocket class does not create the 
 * underlying file object *until* bind() or connect() are called.
 * @author David Lapsley (dlapsley@haystack.mit.edu)
 */
class FileSocket: public Socket {
  public:
	 /** Default constructor.
	  */
  FileSocket();
	 /** Constructor.
	  * The constructor opens the file with the supplied name and opens
	  * it in the read/write mode.
	  * @param mtu maximum block size to read/write.
	  */
  FileSocket(const int& mtu);
	 /** Constructor.
	  * The constructor opens the file with the supplied name and opens
	  * it in the read/write mode.
	  * @param fd file descriptor.
	  * @param mtu maximum block size to read/write.
	  */
  FileSocket(const int& fd, const int& mtu);
	 /** Destructor. */
  ~FileSocket();
	 /** Socket API method.
	  * Attempt to receive data into supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes read.
	  */
  int recv(u_int8_t* b, const int& n);
	 /** Socket API method.
	  * Attempt to send data in supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes sent.
	  */
  int send(u_int8_t* b, const int& n);
  int sendto(const struct sockaddr* to, const int tolen, char* s, const int n){
    cerr << "Warning FILESocket::sendto(): Do Nothing in FILE" << endl; 
    return -1;
  }
  int recvfrom(struct sockaddr& from, int& fromlen, char* s, const int& n){
    cerr << "Warning FILESocket::recvfrom(): Do Nothing in FILE" << endl; 
    return -1;
  }
  /** Socket API method.
   * Binds socket to supplied address.
   * @param addr address to bind to.
   * @param port unused. 
   * @return 0: on success, -1 on failure.
   */
  int bind(const string& addr, const int& port=0);
	 /** Socket API method.
	  * Prepare a socket for accepting incoming connections.
	  * @return 0: on success, -1 on failure.
	  */
  int listen();
	 /** Socket API method.
	  * Wait for incoming connections on a socket. 
	  * @return -1: on failure, positive integer on success. On succes
	  *         the return value is a valid socket descriptor.
	  */
  int accept();
	 /** Socket API method.
	  * Attempt to establish a connection to the supplied address.
	  * @param addr address to connect to.
	  * @param port port to connect to.
	  * @return 0: on success, -1 on failure.
	  */
  int connect(const string& addr, const int& port);
	 /** FileSocket connection termination.
	  * Closes the file.
	  * @return 0: on success, -1 on failure.
	  */
  int close();
	 /** FileSocket connection termination.
	  * Closes the file.
	  * @return 0: on success, -1 on failure.
	  */
  int shutdown(const int& howto);
      //  Read open Procedure 
  int rd_open(const string &name);
      //  Write open Procedure 
  int wr_open(const string &name);
  void Print(FILE* fp=stderr){;}
  void Error(FILE* fp=stderr){;}
  int select(const int rwe, double to){return 0;}
  

};

// Read open prodedure 
inline
int FileSocket::rd_open(const string &name){
  _sockd = open(name.c_str(), O_RDONLY | O_LARGEFILE );
  if (_sockd>0)     return(0);
  
  syslog(LOG_ERR, "FileSocket::rd_open(): open error: %s", strerror(errno));
  return(-1);
}

// Write open prodedure without append
inline 
int FileSocket::wr_open(const string &name){
  //  _sockd=open(name.c_str(), O_WRONLY | O_TRUNC | O_CREAT | O_APPEND, 
  _sockd=open(name.c_str(), O_WRONLY | O_TRUNC | O_CREAT , 
	      S_IRUSR | S_IWUSR | O_LARGEFILE );
  if (_sockd>0)     return(0);
  
  syslog(LOG_ERR, "FileSocket::wr_open(): open error: %s", strerror(errno));
  return(-1);
}

#endif


