/*
 * Copyright (c) 2003 MIT, Haystack Observatory
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction,including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 */

/*
 * Author:	Trevor Cappallo
 */

#include <stdio.h>
#include <stdlib.h>
#include <K5.h>
#include <syslog.h>
#include <sys/ioctl.h>

K5::K5(const bool udp, const int& mjd, const string& remote_ip, const int& data_port):
		  aaaa(0),
		  dddd(0),
		  datamask(0),
		  outmask(0),
		  xxmask(0),
		  data_d_k(0),
		  xxp(0),
		  _tx_buf_idx(0),
		  _mjd(mjd),
		  _udp(udp)
{

	 init_parity(parity_table);
	 init_bitreverse(bitreverse_table);

	 // Incoming Connection setup.
	 // 1. Setup control port for receiving incoming connections.
	 cout << "K5: Binding to 0.0.0.0/" << data_port << endl;
	 _accept_sock.bind("0.0.0.0", data_port);
	 _accept_sock.listen();
	 cout << "K5: Accepting incoming ctrl connections\n";
	 // 2. Accept incoming control connections.        
	 for (int i=0; i<4; ++i) {
		  fd[i]=_accept_sock.accept();
		  if (fd[i]<=0) {
				syslog(LOG_ERR, "Invalid accept\n");
				return;
		  }
		  int rcvbuf=4000000;
		  if (setsockopt(fd[i], SOL_SOCKET, SO_RCVBUF, &rcvbuf, sizeof(int))!=0) {
				syslog(LOG_ERR, "K5: unable to set socket rcv buf size.");
		  }	
		  cout << "K5: Got incoming data connection " << i << endl;
	 }

	 // 3. Create outgoing connection.
	 if (udp) {
		  cout << "K5: Creating outgoing UDP connection\n";
		  _udp_out_sock.connect(remote_ip, data_port);
		  fdout=_udp_out_sock.get_sockd();
	 } else {
		  cout << "K5: Creating outgoing TCP connection\n";
		  _tcp_out_sock.connect(remote_ip, data_port);
		  fdout=_tcp_out_sock.get_sockd();
	 }
	 cout << "K5: completed connection setup.\n";
}

K5::~K5()
{
	 for (int i=0; i<4; ++i)
		  close(fd[i]);
}

int K5::slow_parity(unsigned char c) 
{
	 int i;
	 int n = 0;
	 for (i = 0; i < 8; i++) {
		  n += !!(c & (1 << i));
	 }
	 n = (n + 1) % 2;
	 return (n);
}

void K5::init_parity(int* table) 
{
	 unsigned char c;
	 for (unsigned int i=0; i<=255; ++i) {
		  c=(unsigned char)i;
		  table[i]=slow_parity(c);
		  printf("table[%d]: %d\n", c, table[c]);
	 }
}

void K5::init_bitreverse(unsigned int *table)
{
	 for (unsigned int i=0; i<=65535; ++i) {
		  unsigned int crc16 = i;
		  crc16 = ((crc16 & 0x8000) >> 15)
				| ((crc16 & 0x4000) >> 13)
				| ((crc16 & 0x2000) >> 11)
				| ((crc16 & 0x1000) >> 9)
				| ((crc16 & 0x0800) >> 7)
				| ((crc16 & 0x0400) >> 5)
				| ((crc16 & 0x0200) >> 3)
				| ((crc16 & 0x0100) >> 1)
				| ((crc16 & 0x0080) << 1)
				| ((crc16 & 0x0040) << 3)
				| ((crc16 & 0x0020) << 5)
				| ((crc16 & 0x0010) << 7)
				| ((crc16 & 0x0008) << 9)
				| ((crc16 & 0x0004) << 11)
				| ((crc16 & 0x0002) << 13)
				| ((crc16 & 0x0001) << 15);
		  table[i]=crc16;
	 }
	 for (int j=1; j<65536; j*=2)
		  printf("bitreverse[%d]: %d\n", j, table[j]);
}

void K5::run() {
  // FILE           *fp[4], *fpOut;
  int channels, freq, quant;
  int day[3] = {0, 0, 0};
  int seconds[5] = {0, 0, 0, 0, 0};
  int secondParts[4] = {0, 0, 0, 0};
  int	 sp, spInc, a, b, i, k, c, m, d;
  long time, n;
  const int files = 4;
  
  /* split day into three digits */
  day[0] = (int) (_mjd / 100);
  day[1] = (int) ((_mjd % 100) / 10);
  day[2] = (int) (_mjd % 10);
  
  /* write NRZ initial word */
  nrzOut = FWORD;
  
  write(fdout, &nrzOut, sizeof(int));
  
  /* loop through input file */
  while ( (unsigned)(n = read(fd[0], data[0], sizeof(int)*2)) 
	  >= 2*sizeof(int) ) {
    for (i = 1; i < files; i++) {
      read (fd[i], data[i], sizeof(int)*2);
    }
    
    /* parse input header */
    
    if (data[0][1] & CHAN)
      channels = 4;
    else
      channels = 1;
    
    quant = 1 << ((data[0][1] & QUANT) >> 22);
    
    freq = (data[0][1] & FREQ) >> 18;
    if (freq == 0)  { freq = 40000; }
    else if (freq == 1)  { freq = 100000; }
    else if (freq == 2)  { freq = 200000; }
    else if (freq == 3)  { freq = 500000; }
    else if (freq < 9)  { freq = 1000000 * (1 << (freq - 4)); }
    else  { 
      /* error: invalid value */ 
      fprintf(stdout, "Invalid value for frequency: %d",
	      freq);
    }
    
    time = data[0][1] & 0x1FFFF;
    if (time >= 86400)  { 
      /* error: invalid value */ 
      fprintf(stdout, "Invalid value for time: %ld",
	      time);
    }
    
    c = 0;
    m = 0;
    sp = 0;                               /* second parts as an integer * 10 */
    spInc = (int) (2000000000 / freq);    /* increment for sp each frame */
    int y=0, z=0;
    const int MAX_ITER=100;
    const int RD_SIZE=sizeof(int)*MAX_SIZE;
    int queued=0;

    int max=0;
    int w;
    for (w=0; w<files; ++w) {
      if (fd[w]>max)
	max=fd[w];
    }
    ++max;		

    /* read data chunk(s) */
    for (i = 0; i < (freq / 32 * channels * quant) / MAX_SIZE; i++) {
      for (a = 0; a < files; a++) {
	u_int8_t* p=(u_int8_t*)&(data[a]);
	int length=0;
	int ret=0, req=0;
	y=0;
	z=0;
	fd_set rfds;
	int u;
	for (u=0; u<files; ++u)
	  FD_SET(fd[u], &rfds);

	if (select(max, &rfds, NULL, NULL, NULL)<0) {
	  syslog(LOG_ERR, "Select() error %s", strerror(errno));
	  break;
	}
	do {
	  req=RD_SIZE-length;
	  do {
	    if (ioctl(fd[a], FIONREAD, &queued)!=0) {
	      syslog(LOG_ERR, "K5: ioctl error %s", strerror(errno));
	      break;
	    }
	    // FIXME cerr << "* "; 
	    if (y++>MAX_ITER)
	      break;
	  } while (queued<req);
	  ret=read(fd[a], p, req);
	  if (ret>0) {
	    length+=ret;
	    if (length<RD_SIZE) {
	      p+=ret;
	    } else if (length>=RD_SIZE) {
	      break;
	    }
	  }
	  if (z++>MAX_ITER) {
	    syslog(LOG_ERR, "K5: read(): max iterations exceeded");
	    return;
	  }
	  // FIXME cerr << ".";
	} while (length<RD_SIZE || (ret<0 && errno==EINTR));
	n = length;
	// FIXME cerr << "K5: Read " << n << " bytes\n";
	// FIXME cerr << " +\n";
	// convert n back to integers.
	n/=4;
      }
      // FIXME cerr << "=";

      /* loop for each word of data */
      for (k = 0; k < n; k++) {
	if (m == 0) {
	  /* write header */
					
	  /* sync words */
	  for (a = 0; a < 4; a++) {
	    for (b = 0; b < 8; b++) {
	      out = FWORD;
	      WRITE_WORD ();
	    }
	    out = 0;
	    WRITE_WORD ();
	  }
					
	  /* JJJ */
	  for (a = 0; a < 3; a++) {
	    if (a == 2) {
	      out = 0;
	      x = (day[0] << 4) + day[1];
	      if (fast_parity(x)) {
		out = FWORD;
	      }
	      WRITE_WORD ();
	    }
						
	    for (b = 3; b >= 0; b--) {
	      out = 0;
	      if (day[a] & (1 << b)) {
		out = FWORD;
	      }
	      WRITE_WORD ();
	    }
	  }
					
	  /* SSSSS */
	  seconds[0] = (int) (time / 10000);
	  seconds[1] = (int) ((time % 10000) / 1000);
	  seconds[2] = (int) ((time % 1000) / 100);
	  seconds[3] = (int) ((time % 100) / 10);
	  seconds[4] = time % 10;

	  for (a = 0; a < 5; a++) {
	    for (b = 3; b >= 0; b--) {
	      out = 0;
	      if (seconds[a] & (1 << b)) {
		out = FWORD;
	      }
	      WRITE_WORD ();
	    }
						
	    if (a == 0) {
	      out = 0;
	      x = (day[2] << 4) + seconds[0];
	      if (fast_parity(x)) {
		out = FWORD;
	      }
	      WRITE_WORD ();
	    }
	    else if ((a == 2) || (a == 4)) {
	      out = 0;
	      x = (seconds[a - 1] << 4) + seconds[a];
	      if (fast_parity(x)) {
		out = FWORD;
	      }
	      WRITE_WORD ();
	    }
	  }

	  /* ssss */
	  secondParts[0] = (int) (sp / 10000);
	  secondParts[1] = (int) ((sp % 10000) / 1000);
	  secondParts[2] = (int) ((sp % 1000) / 100);
	  secondParts[3] = (int) ((sp % 100) / 10);

	  for (a = 0; a < 4; a++) {
	    for (b = 3; b >= 0; b--) {
	      out = 0;
	      if (secondParts[a] & (1 << b)) {
		out = FWORD;
	      }
	      WRITE_WORD ();
	    }

	    if ((a == 1) || (a == 3)) {
	      out = 0;
	      x = (secondParts[a - 1] << 4) + secondParts[a];
	      if (fast_parity(x)) {
		out = FWORD;
	      }
	      WRITE_WORD ();
	    }
	  }	
					
	  sp += spInc;
					
	  /* CRC-16 */
	  iData[0] = (day[0] << 4) + day[1];
	  iData[1] = (day[2] << 4) + seconds[0];
	  iData[2] = (seconds[1] << 4) + seconds[2];
	  iData[3] = (seconds[3] << 4) + seconds[4];
	  iData[4] = (secondParts[0] << 4) + secondParts[1];
	  iData[5] = (secondParts[2] << 4) + secondParts[3];

	  unsigned int fwd_crc16 = crcc(iData, 48, 040003, 16);
	  crc16 = bitreverse(fwd_crc16);

	  /* write CRC with fast_parity */
	  x = 0;
	  for (a = 15; a >= 8; a--) {
	    out = 0;
	    if (crc16 & (1 << a)) {
	      out = FWORD;
	      x += (1 << (a - 8));
	    }
	    WRITE_WORD ();
	  }
	  out = 0;
	  if (fast_parity(x)) {
	    out = FWORD;
	  }
	  WRITE_WORD ();
					
	  x = 0;
	  for (a = 7; a >= 0; a--) {
	    out = 0;
	    if (crc16 & (1 << a)) {
	      out = FWORD;
	      x += (1 << a);
	    }
	    WRITE_WORD ();
	  }
	  out = 0;
	  if (fast_parity(x)) {
	    out = FWORD;
	  }
	  WRITE_WORD ();
	}
				
	out = 0;
	if (channels != 4) {
	  /* error, this program only plans on 4 */
	  fprintf(stdout, "Program only supports 4 channels\n");
	}

	for (a = 0; a < 32; a++) {
	  xx[a] = 0;
	}

#if 0
	/* for 1-bit sampling quantization */
	if (quant == 1) {
	  /* there are 8 sets of 4 channels with 1 bit samples */
	  for (a = 0; a < 8; a++) {
	    out = 0;
	    /* copy data from each file into part of the word */
	    for (d = 0; d < files; d++) {
	      for (b = 0; b < 4; b++) {
		if (data[d][k] & (1 << (4*a + b))) {
		  out += (1 << (b + 4*d));
		  xx[b + 4*d] += (1 << a);
		}
	      }
	    }
	    WRITE_WORD ();
	  }

	  /* write parity word*/
	  out = 0;
	  for (a = 0; a < 32; a++) {
	    if (fast_parity(xx[a])) {
	      out += (1 << a);
	    }
	  }
	  WRITE_WORD ();
	} else  { 
	  /* error, program only plans on 1 bit sampling */ 
	  fprintf(stdout, "Program only plans on 1 bit sampling\n"); 
	}
#else
	/* for 1-bit sampling quantization */
	/* See OPTIM_DECLS above */
	if (quant == 1) {
	  /* there are 8 sets of 4 channels with 1 bit samples */
	  for (a = 0; a < 8; a++) {
	    out = 0;
	    aaaa=4*a;
	    xxmask=(1<<a);
	    datamask=(1<<aaaa);
	    /* copy data from each file into part of the word */
	    for (d = 0; d < files; d++) {
	      dddd=4*d;
	      outmask=(1<<dddd);
	      data_d_k=data[d][k];
	      xxp=&xx[dddd];
	      /* Unrolled loop */
	      if (data_d_k & datamask) {
		out += (outmask);
		(*xxp) += xxmask;
	      }
	      ++xxp;
	      if (data_d_k & (datamask<<1)) {
		out += (outmask<<1);
		(*xxp) += xxmask;
	      }
	      ++xxp;
	      if (data_d_k & (datamask<<2)) {
		out += (outmask<<2);
		(*xxp) += xxmask;
	      }
	      ++xxp;
	      if (data_d_k & (datamask<<3)) {
		out += (outmask<<3);
		(*xxp) += xxmask;
	      }
	      ++xxp;
	    }
	    WRITE_WORD ();
	  }

	  /* write parity word*/
	  out = 0;
	  for (a = 0; a < 32; a++) {
	    if (fast_parity(xx[a])) {
	      out += (1 << a);
	    }
	  }
	  WRITE_WORD ();
	} else  { 
	  /* error, program only plans on 1 bit sampling */ 
	  printf("Program only plans on 1 bit sampling\n"); 
	}
#endif
				
	sp = sp % 100000;
	m++;
				
	if (m == 2500) {
	  m = 0;

	  /* write aux data */
					
	  for (a = 0; a < 8; a++) {
	    for (b = 0; b < 8; b++) {
	      out = 0;
	      WRITE_WORD ();
	    }
	    out = FWORD;
	    WRITE_WORD ();
	  }
	}
      }
			
      c++;
    }

    /* print some junk */
    printf ("Data read: %d\n", c * MAX_SIZE);
    printf ("Channels: %d\n", channels);
    printf ("Quantization: %d bit(s)\n", quant);
    printf ("Frequency: %d Hz\n", freq);
    printf ("Time: %ld secs after midnight\n", time);
    printf ("\n");
  }
}

