/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2003 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */

#include <unistd.h>
#include <RTCPClientSocket.h>

RTCPClientSocket::RTCPClientSocket(Socket& s,
											  RTPSession& rtp_session,
											  const int& mtu):
		  RTCPSocket(s, rtp_session, mtu)
{
}


RTCPClientSocket::RTCPClientSocket(const RTCPClientSocket& r):
		  RTCPSocket(r)
{
}

RTCPClientSocket::~RTCPClientSocket()
{
}

int RTCPClientSocket::recv_proc()
{
  int ret=RTCPSocket::recv(_buf, _mtu);
  if (ret<_mtu) {
    syslog(LOG_ERR, "RTCPClientSocket: recv() error: %s", strerror(errno));
  }
  RTCPCommon* h=(RTCPCommon*)_buf;
  RTCPPacket* hh=0;
  u_int32_t ntp_sec=0, ntp_frac=0;
  switch (h->_pt) {
  case RTCP_SR:
    syslog(LOG_DEBUG, "RTCPClientSocket: Received SR!");
    if (gettimeofday(&_rtp_session._r._lsrr, NULL)!=0)
      syslog(LOG_ERR, "RTCPClientSocket: gettimeofday error: %s", 
	     strerror(errno));
    hh=(RTCPPacket*)_buf;
    _rtp_session._ssrc=ntohl(hh->r.sr._ssrc);
    ntp_sec=ntohl(hh->r.sr._ntp_sec);
    ntp_frac=ntohl(hh->r.sr._ntp_frac);
    _rtp_session._r._ntp_sec=ntp_sec;
    _rtp_session._r._ntp_frac=ntp_frac;
    _rtp_session._r._rtp_ts=ntohl(hh->r.sr._rtp_ts);
    _rtp_session._s._psent=ntohl(hh->r.sr._psent);
    _rtp_session._s._osent=ntohl(hh->r.sr._osent);
    _rtp_session._r._lsr=(ntp_sec<<16)+(ntp_frac>>16);
#ifdef PRINT
    _rtp_session.print_sr((u_int8_t*)rtcp_buf[0].iov_base);
#endif // PRINT
    break;
  case RTCP_SDES:
    syslog(LOG_NOTICE, "Received SDES!");
    break;
  case RTCP_BYE:
    syslog(LOG_NOTICE, "Received BYE!");
    break;
  case RTCP_APP:
    syslog(LOG_NOTICE, "Received APP!");
    break;
  default:
    syslog(LOG_ERR, "Received unknown RTCP packet!");
    break;
  }
  return(0);
}

int RTCPClientSocket::send_proc()
{
  unsigned int count=1; 
  
  // <Ref: RFC3550:Appendix A.3.>
  u_int32_t extended_max=_rtp_session._r._cycles+
    _rtp_session._r._max_seq;
  u_int32_t expected=extended_max-_rtp_session._r._base_seq+1;
	 int32_t lost=expected-_rtp_session._r._received;
	 lost=(lost>0x7fffff)?0x7fffff:lost;
	 lost=(lost<0)?0x800000:lost;
	 _rtp_session._r._lost=lost;
	 u_int32_t expected_interval=expected-
		  _rtp_session._r._expected_prior;
	 _rtp_session._r._expected_prior=expected;
	 u_int32_t received_interval=_rtp_session._r._received-
		  _rtp_session._r._received_prior;
	 _rtp_session._r._received_prior=_rtp_session._r._received;
	 u_int32_t lost_interval=expected_interval-received_interval;
	 if (expected_interval==0 || lost_interval<=0) 
		  _rtp_session._r._fraction=0;
	 else 
		  _rtp_session._r._fraction=(lost_interval<<8)/
				expected_interval;
	 double delta = received_interval*_mtu;
	 double throughput=8.0*delta/_rtcp_duration.elapsed();
	 cerr << "throughput " << throughput << endl;

	 syslog(LOG_DEBUG, "FRACTION: %u", _rtp_session._r._fraction);
	 syslog(LOG_DEBUG, "LOST_INTERVAL: %u", lost_interval);
	 syslog(LOG_DEBUG, "EXPECTED_INTERVAL: %u", expected_interval);
	 // </Ref: RFC3550:Appendix A.3.>
	 struct timeval now, diff;
	 if (gettimeofday(&now, NULL)!=0) 
		  syslog(LOG_ERR, "RTCPClientSocket: gettimeofday error: %s", 
					strerror(errno));
	 timersub(&now, &_rtp_session._r._lsrr, &diff);
	 double ddiff=diff.tv_sec+(diff.tv_usec/1e6); 
	 ddiff=ddiff*65536;
	 _rtp_session._r._dlsr=int(ddiff);
	 _rtcp_duration.stop();
	 RTCPPacket* hh;
	 hh=(RTCPPacket*)build_rr(_buf, 
									  _mtu,
									  count, 
									  _rtp_session._ssrc, 
									  _rtp_session._r._fraction, 
									  _rtp_session._r._lost, 
									  extended_max,
									  (u_int32_t)_rtp_session._r._jitter, 
									  _rtp_session._r._lsr, 
									  _rtp_session._r._dlsr);
	 const u_int16_t rr=0;
	 const u_int16_t pid=_rtp_session._r._max_seq;
	 const u_int16_t blp=0;
	 const u_int16_t blp_sz=1;
	 build_ack((u_int8_t*)hh, 
				  _mtu,
				  _rtp_session._ssrc,
				  _rtp_session._ssrc,
				  &rr,
				  &pid,
				  &blp,
				  blp_sz);
#ifdef PRINT
	 print_rr((u_int8_t*)_buf);
	 print_ack((u_int8_t*)hh);
#endif // PRINT
	 unsigned int r=0;
	 int ret=RTCPSocket::send(_buf, _mtu);
	 if (ret<0) {
		  syslog(LOG_ERR, "RTCPClientSocket: send() error: %s", strerror(errno));
	 }
	 // Update average packet size.
	 _rtp_session._r._avg_rtcp_size = (1/16)*(ret+8+20)+                                                
		  (15/16) * _rtp_session._r._avg_rtcp_size;
	 syslog(LOG_DEBUG, "RTCPClientSocket: bytes_sent==%d", r);
	 _rtcp_duration.start();
	 return(ret);
}



