/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2003 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */

#ifndef RTCPSOCKET_H
#define RTCPSOCKET_H

#include <common.h>
#include <TCPSocket.h>
#include <string>
#include <iostream>
#include <RTPSession.h>
#include <Utils.h>

/** RTCPSocket class.
 * This class provides the base class for RTCP sockets. It inherits from TCPSocket.
 
 */

class RTCPSocket : public TCPSocket {
 protected:
	 /** RTCP Parameter.
	  * Reference to a RTPSession object that is used to store RTCP state.
	  */
  RTPSession& _rtp_session;
	 /** Network.
	  * Buffer for storing RTCP packets.
	  */
  u_int8_t* _buf;
	 /** Timer.
	  * Used to time the interval between RTCP transmissions.
	  */
  Timer _rtcp_duration;
  public:
	 /** Constructor.
	  * This calls the base constructor to create an RTCPSocket.
	  */
  //  RTCPSocket() :TCPSocket(){;}
	 /** Constructor.
	  * This constructor creates a socket and sets the internal MTU member
	  * to be equal to the supplied parameter.
	  * @param sock reference to underlying transport socket.
	  * @param rtp_session a reference to an RTPSession that will be used to
	  * store the protocol state.	 
	  * @param mtu size of packets to be transmitted.
	  */
  RTCPSocket(Socket& sock,
	     RTPSession& rtp_session,
	     const int& mtu);
	 /** Copy constructor.
	  * Copies values of other VRClientSocket into this object.
	  * @param s VRClientSocket to copy.
	  */
  RTCPSocket(const RTCPSocket &r);
	 /** Destructor. */	  
  virtual ~RTCPSocket();
  /** Processing method.
	  * Recv RTCP compound packet from client side, process and store
	  * result in RTPSession structure.
	  */
  virtual  int recv_proc() = 0;
	 /** Processing method.
	  * Send RTCP compound packet to client side, process and store
	  * result in RTPSession structure.
	  */
  virtual int send_proc() = 0;
	 /** Socket API method.
	  * Attempt to receive data into supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes read.
	  */
//   inline int recv(u_int8_t* b, const int& n) {
//     return(TCPSocket::recv(b, n));
//   }
         /** Socket API method.
	  * Attempt to send data in supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes sent.
	  */
//   inline int send(u_int8_t* b, const int& n) {
//     return(_sock.send(b, n));
//   }
	 /** Transport function.
	  * Builds an RTCP sender report packet.
	  * @param sb array to store data header in.
	  * @param sz size of sb array.
	  * @param count count of receiver reports contained in this RR.
	  * @param ssrc synchronization source.
	  * @param ntp_sec NTP timestamp value.
	  * @param ntp_frac NTP timestamp value.
	  * @param rtp_ts RTP Timestamp.
	  * @param psent packets sent.
	  * @param osent octets sent.
	  * @return pointer to one past end of buffer.
	  */
  u_int8_t* build_sr(u_int8_t* sb, const int& sz, const u_int8_t count, 
		     const u_int32_t& ssrc, const u_int32_t& ntp_sec, 
		     const u_int32_t& ntp_frac, const u_int32_t& rtp_ts, 
		     const u_int32_t& psent, const u_int32_t& osent);
	 /** Transport function.
	  * Builds an RTCP BYE packet.
	  * @param sb array to store data header in.
	  * @param sz size of sb array.
	  * @param src_count number of sources included in this packet.
	  * @param ssrcs array of source ssrc's.
	  */
  u_int8_t* build_bye(u_int8_t* sb, const int& sz,
		      const u_int8_t src_count,
		      const u_int32_t* ssrcs);
	 /** Transport function.
	  * Builds an RTCP receiver report packet.
	  * @param sb array to store data header in.
	  * @param sz size of sb array.
	  * @param count count of receiver reports contained in this RR.
	  * @param ssrc synchronization source.
	  * @param fraction  fraction of packets lost since last SR received.
	  * @param lost total number of packets lost since last SR received.
	  * @param last_seq sequence number of last data packet received.
	  * @param jitter  running average jitter.
	  * @param lsr last sender report.
	  * @param dlsr delay since last sender report.
	  * @return pointer to one past end of buffer.
	  */
  u_int8_t* build_rr(u_int8_t* sb, const int& sz, const u_int8_t count, 
		     const u_int32_t& ssrc, const u_int32_t& fraction, 
		     const u_int32_t& lost, const u_int32_t& last_seq, 
		     const u_int32_t& jitter, const u_int32_t& lsr, 
		     const u_int32_t& dlsr);
	 /** Transport function.
	  * Builds an RTCP feedback packet.
	  * @param sb array to store data header in.
	  * @param sz size of sb array.
	  * @param sender_ssrc SSRC of packet sender.
	  * @param src_ssrc SSRC of media source.
	  * @param pid Packet ID. This identifies the packet that is being NACK-ed.
	  * @param Bit-mask of following lost packets. 
	  * bitmask of following lost packets (BLP): 16 bits  
	  * The BLP allows for reporting losses of any of the 16 RTP 
	  * packets immediately following the RTP packet indicated by the 
	  * PID.  The BLP's definition is identical to that given in [6].  
	  * Denoting the BLP's least significant bit as bit 1, and its most 
	  * significant bit as bit 16, then bit i of the bit mask is set to 
	  * 1 if the receiver has not received RTP packet number (PID+i) 
	  * (modulo 2^16) and indicates this packet is lost; bit i is set 
	  * to 0 otherwise.  Note that the sender MUST NOT assume that a 
	  * receiver has received a packet because its bit mask was set to 
	  * 0.   For example, the least significant bit of the BLP would be 
	  * set to 1 if the packet corresponding to the PID and the 
	  * following packet have been lost.  However, the sender cannot 
	  * infer that packets PID+2 through PID+16 have been received 
	  * simply because bits 2 through 15 of the BLP are 0; all the 
	  * sender knows is that the receiver has not reported them as lost 
	  * at this time(from draft-ietf-avt-rtcp-feedback-08.txt).
	  * @param blp_sz N#include <RTPSession.h>umber of blps.
	  */
  u_int8_t* build_nack(u_int8_t* sb, const int& sz,
		       const u_int32_t& sender_ssrc,
		       const u_int32_t& src_ssrc,
		       const u_int16_t* pid,
		       const u_int16_t* blp,
		       int blp_sz);
	 /** Transport function.
	  * Builds an RTCP feedback packet.
	  * @param sb array to store data header in.
	  * @param sz size of sb array.
	  * @param sender_ssrc SSRC of packet sender.
	  * @param src_ssrc SSRC of media source.
	  * @param pid Packet ID of packet being ACK-ed.
	  * @param r Range of ACKs. 
	  * @param blp  Bit mask of Lost Packets/# Packets
	  * @param blp_sz Number of blps.
	  */
  u_int8_t* build_ack(u_int8_t* sb, const int& sz,
		      const u_int32_t& sender_ssrc,
		      const u_int32_t& src_ssrc,
		      const u_int16_t* r,
		      const u_int16_t* pid,
		      const u_int16_t* blp,
		      int blp_sz);
	 /**  Debug function.
	  *  Prints an sr packet to debug stream.
	  * @param sb buffer that contains RR.
	  */
  void print_sr(u_int8_t* sb, FILE* fp=stderr);
	 /**  Debug function.
	  * @param sb buffer that contains RR.
	  *  Prints an rr packet to debug stream.
	  */
  void print_rr(u_int8_t* sb, FILE* fp=stderr);
	 /**  Debug function.
	  *  Prints a nack packet to debug stream.
	  * @param sb buffer that contains RR.
	  */
  void print_nack(u_int8_t* sb, FILE* fp=stderr);
	 /**  Debug function.
	  *  Prints a ack packet to debug stream.
	  * @param sb buffer that contains RR.
	  */
  void print_ack(u_int8_t* sb, FILE* fp=stderr);
};

ostream& operator<<(ostream& os, const RTCPSocket& r);

inline 
u_int8_t* RTCPSocket::build_sr(u_int8_t* sb, const int& sz, const u_int8_t count, 
			       const u_int32_t& ssrc, const u_int32_t& ntp_sec, 
			       const u_int32_t& ntp_frac, const u_int32_t& rtp_ts, 
			       const u_int32_t& psent, const u_int32_t& osent) {
  RTCPPacket* h=(RTCPPacket*)sb;
  h->h._common._v=0x2;
  h->h._common._p=0;
  h->h._common._pt=RTCP_SR;
  h->h._common._length=htons(7);
  h->r.sr._ssrc=htonl(ssrc);
  h->r.sr._ntp_sec=htonl(ntp_sec);
  h->r.sr._ntp_frac=htonl(ntp_frac);
  h->r.sr._rtp_ts=htonl(rtp_ts);
  h->r.sr._psent=htonl(psent);
  h->r.sr._osent=htonl(osent);
  sb+=28;
  return(sb);
}

inline 
u_int8_t* RTCPSocket::build_bye(u_int8_t* sb, const int& sz,
				const u_int8_t src_count,
				const u_int32_t* ssrcs)
{
  RTCPPacket* h=(RTCPPacket*)sb;
  h->h._common._v=0x2;
  h->h._common._p=0;
  h->h._common._pt=RTCP_BYE;
  h->h._common._count=src_count;
  h->h._common._length=htons(src_count);
  for (int i=0; i<src_count; ++i) {
    h->r.bye._src[i]=htonl(ssrcs[i]);
  }
  sb+=(src_count+1)*4;
  return(sb);
}

inline 
u_int8_t* RTCPSocket::build_rr(u_int8_t* sb, const int& sz, const u_int8_t count, 
			       const u_int32_t& ssrc, const u_int32_t& fraction, 
			       const u_int32_t& lost, const u_int32_t& last_seq, 
			       const u_int32_t& jitter, const u_int32_t& lsr, 
			       const u_int32_t& dlsr) {
  RTCPPacket* h=(RTCPPacket*)sb;
  h->h._common._v=0x2;
  h->h._common._p=0;
  h->h._common._pt=RTCP_RR;
  h->h._common._length=htons(6);
  h->r.rr._rr[0]._ssrc=htonl(ssrc);
  h->r.rr._rr[0]._fraction=0xff&fraction;
  h->r.rr._rr[0]._lost=htonl(0xffffff&lost)>>8;
  h->r.rr._rr[0]._last_seq=htonl(last_seq);
  h->r.rr._rr[0]._jitter=htonl(jitter);
  h->r.rr._rr[0]._lsr=htonl(lsr);
  h->r.rr._rr[0]._dlsr=htonl(dlsr);
  sb+=28;
  return(sb);
}

inline 
u_int8_t* RTCPSocket::build_nack(u_int8_t* sb, const int& sz,
				 const u_int32_t& sender_ssrc,
				 const u_int32_t& src_ssrc,
				 const u_int16_t* pid,
				 const u_int16_t* blp,
				 int blp_sz) {
  RTCPPacket* h=(RTCPPacket*)sb;
  h->h._common._v=0x2;
  h->h._common._p=0;
  h->h._common._pt=RTCP_FB;
  h->h._common._count=FB_NACK;
  h->h._common._length=htons(3+blp_sz);
  h->r._fb._sender_ssrc=htonl(sender_ssrc);
  h->r._fb._src_ssrc=htonl(src_ssrc);
  for (int i=0; i<blp_sz; ++i) {
    h->r._fb._fci[i]._ack._pid=htons(pid[i]);
    h->r._fb._fci[i]._ack._blp=htons(blp[i]);
  }
  sb+=(3+blp_sz)*4;
  return(sb);
}

inline 
u_int8_t* RTCPSocket::build_ack(u_int8_t* sb, const int& sz,
				const u_int32_t& sender_ssrc,
				const u_int32_t& src_ssrc,
				const u_int16_t* r,
				const u_int16_t* pid,
				const u_int16_t* blp,
				int blp_sz) {
  RTCPPacket* h=(RTCPPacket*)sb;
  h->h._common._v=0x2;
  h->h._common._p=0;
  h->h._common._pt=RTCP_FB;
  h->h._common._count=FB_ACK;
  h->h._common._length=htons(3+blp_sz);
  h->r._fb._sender_ssrc=htonl(sender_ssrc);
  h->r._fb._src_ssrc=htonl(src_ssrc);
  for (int i=0; i<blp_sz; ++i) {
    h->r._fb._fci[i]._ack._r=htons(r[i]);
    h->r._fb._fci[i]._ack._pid=htons(pid[i]);
    h->r._fb._fci[i]._ack._blp=htons(blp[i]);
  }
  sb+=(3+blp_sz)*4;
  return(sb);
}

#endif


