/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2004 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */


#ifndef SOCKET_H
#define SOCKET_H

#include <common.h>
#include <string>
#include <iostream>
#ifdef LINUX
#include <unistd.h>
#endif // LINUX

extern "C"{
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <arpa/inet.h>
}

#include <Utils.h>
#define IPTOS_QBSS 0x20

using namespace std;

/** Socket module
 * @doc This module implements base level socket functionality. It
 * provides a number of key classes that are utilized by derived classes.
 * @author David Lapsley (dlapsley@haystack.mit.edu)
 * 
 */

/** Base level socket class.
 * All other socket classes are derived from this one. This class inherits
 * from the Object class.
 */
class Socket {
  protected:
	 /** @doc Encapsulated socket descriptor.  */
    int _sockd;
	 /** @doc Maximum Transmission Unit for this socket.  */
    int _mtu;
	 /** @doc Total number of bytes sent during the lifetime of this socket.  */
    long long _bytes_sent;
	 /** @doc Total number of bytes received during the lifetime of this 
	  * socket.  
	  */
    long long _bytes_rcvd;
	 /** @doc Total number of send calls made during the lifetime of this 
	  * socket.
	  */
    long long _send_calls;
	 /** @doc Total number of recv calls made during the lifetime of this socket.
	  */
    long long _recv_calls;
	 /** @doc Timer for measuring lifetime of socket */
    Timer _age;
    /** Timeout parameter 
	(default 10msec)
    */
    double _timeout;
  public:
	 /** @doc Constant flag value used in set_blocking().  */
    static const int BLOCKING;
	 /** @doc Constant flag value used in set_blocking().  */
    static const int NONBLOCKING;
	 /** @doc Constant flag value used in select() */
    static const int SEL_READ;
	 /** @doc Constant flag value used in select() */
    static const int SEL_WRITE;
	 /** @doc Constant flag value used in select() */
    static const int SEL_EXCEPT;
 	 /** Default constructor. 
	  * Maximum Transmission Unit is set to 1024 by 
	  * default. Socket type is set to SOCK_STREAM.
	  */
    Socket();
	 /** Constructor.
	  * @param m MTU of this socket.
	  */
    Socket(const int m);
	 /** Constructor. 
	  * This constructor does NOT create a socket. It assigns 
	  * the internal socket descriptor member (_sockd) to the supplied 
	  * socket descriptor.
	  * @param s pre-opened socket descriptor (e.g. as returned from 
	  *        C socket() or accept() call). 
	  * @param m socket Maximum Transmission Unit.
	  * @return none
	  * @doc constructs a Socket object.
	  */
    Socket(const int& s, const int m);
	 /** Copy constructor.
	  */
    Socket(const Socket &s);
	 /** Destructor. Important! The destructor does not close() the socket. 
	  * This is because there are situations when it is necessary to have an
	  * opened socket survive the Socket object that created it.
	  * Users of the Socket class are responsible for closing the 
	  * any sockets created using the close() or shutdown() methods.
	  */
    virtual ~Socket();
	 /** Socket API method.
	  * Binds socket to supplied address.
	  * @param ip IP address to bind to.
	  * @param port port to bind to.
	  * @return 0: on success, -1 on failure.
	  */
    virtual int bind(const string& ip, const int& port) = 0;
	 /** Socket API method.
	  * Prepare a socket for accepting incoming connections.
	  * @return 0: on success, -1 on failure.
	  */
    virtual int listen() = 0;
	 /** Socket API method.
	  * Wait for incoming connections on a socket. 
	  * @return -1: on failure, positive integer on success. On succes
	  *         the return value is a valid socket descriptor.
	  * @todo resolve with UDT2.
	  */
    virtual int accept() = 0;
	 /** Socket API method.
	  * Attempt to establish a connection to the supplied address.
	  * @param ip IP addr
	  * @param port port to connect to.
	  * @return 0: on success, -1 on failure.
	  */
    virtual int connect(const string& ip, const int& port) = 0;
	 /** Socket API method.
	  * Attempt to receive data into supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes read.
	  */
    virtual int recv(u_int8_t* b, const int& n) = 0;
    /** Socket API method.
	  * Attempt to recv data from supplied buffer and store the destination.
	  * @param from Place to store address of sender.
	  * @param fromlen Length of from address.
	  * @param s socket buffer to store data in.
	  * @param n number of bytes to attempt to recv.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes received.
	  */
    virtual int recvfrom(struct sockaddr& from, int& fromlen, char* s, 
			 const int& n)=0;
	 /** Socket API method.
	  * Attempt to send data in supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes sent.
	  */
    virtual int send(u_int8_t* b, const int& n) = 0;
    /** Socket API meathod for consistency with UDP
     *  Only effective for UDP, but do nothing on TCP   
     */
    virtual int sendto(const struct sockaddr* to, const int tolen, 
		       char* s, const int n)=0;
	 /** Informational method.
	  * @param sd Socket descriptor.
	  * @return IP address of remote point.
	  */
    string get_peer_ip(const int& sd);
	 /** Access method.
	  * @return value of sockd.  
	  */
    int get_sockd() const { return(_sockd); }
	 /** Access method.
 	  * @return value of type.  
	  */
    int get_type() const;
	 /**  Access method.
	  * @return value of mtu.  
	  */
    int get_mtu() const;
	 /** Access method.
	  * @return value of bytes sent.  
	  */
    void set_mtu(const int );
	 /** Access method.
	  * @return value of bytes sent.  
	  */
    long long get_bytes_sent() const;
	 /** Access method.
	  * @return value of bytes received.  
	  */
    long long get_bytes_rcvd() const;
	 /** Access method.
	  * @return value of send calls.  
	  */
    int get_send_calls() const;
	 /** Access method.
	  * @return value of recv calls.  
	  */
    int get_recv_calls() const;
	 /** Access method.
	  * @doc set SO_REUSEADDR socket option
	  * @param yes 0: allow address reuse, 1: disallow address reuse.
	  * @return 0: on success, -1: on failure.
	  */
    int set_so_reuseaddr(const int yes);
	 /** Access method.
	  * @return value of so_reuseaddr.  
	  */
    int get_so_reuseaddr() const;
	 /** Access method.
	  * @doc set SO_LINGER socket option
	  * @param onoff 0: disable linger option, 1: enable linger option.
	  * @param time if linger enabled, how long to linger for. 
	  * @return 0: on success, -1: on failure.
	  */
    int set_so_linger(const int& onoff, const int& time);
	 /** Access method.
	  * @return value of so_linger.  
	  */
    const struct linger get_so_linger() const;
	 /** Access method.
	  * @doc set SO_RCVBUF socket option
	  * @param sz size to set so_rcvbuf.
	  * @return 0: on success, -1: on failure.
	  */
    int set_so_rcvbuf(const int& sz);
	 /** Access method.
	  * @return value of so_rcvbuf.  
	  */
    int get_so_rcvbuf() const;
	 /** @doc set SO_SNDBUF socket option
	  * @param sz size to set so_sndbuf
	  * @return 0: on success, -1: on failure.
	  */
    int set_so_sndbuf(const int & sz);
	 /** Access method.
	  * @return value of so_sndbuf.  
	  */
    int get_so_sndbuf() const;
	 /** Access method.
	  * @doc set SO_RCVLOWAT socket option
	  * @param m value to set rcv_lowat
	  * @return 0: on success, -1: on failure.
	  */
    int set_so_rcvlowat(const int& m);
	 /** Access method.
	  * @return value of rcv_lowat.  
	  */
    int get_so_rcvlowat() const;
	 /** @doc set SO_SNDLOWAT socket option
	  * @param m size to set so_sndlowat
	  * @return 0: on success, -1: on failure.
	  */
    int set_so_sndlowat(const int& m);	
	 /** Access method.
	  * @return value of so_sndlowat.  
	  */
    int get_so_sndlowat() const;	
	 /** Access method.
	  * @doc set TCP_NODELAY socket option
	  * @param n value to set nodelay to. 0: allow delayed sending, 1: disallow
	  * 	   delayed sending.
	  * @return 0: on success, -1: on failure.
	  */
    int set_tcp_nodelay(const int& n);	
	 /** Accessm method.
	  * @return value of tcp_nodelay.  
	  */
    int get_tcp_nodelay() const;	
	 /** Access method.
	  * @doc set O_NONBLOCK socket option
	  * @param b blocking mode: Socket::blocking: node is blocking (default),
	  * Socket::nonblocking: node is non-blocking.
	  * @return 0: on success, -1: on failure.
	  */
    int set_blocking(const int& b);
	 /** Access method.
	  * @doc get current blocking state of the socket
	  * @return current socket blocking mode.
	  */
    int get_blocking() const;
	 /** Access method.
	  *  @doc set IP_TOS (DSCP) socket option. Note that IP_TOS and IP_DSCP
	  *  occupy the same byte in the IP header, but have different 
	  * interpretations. IP_DSCP does not use the 2 least significant bits.
	  * @return 0: on success, -1: on failure.
	  */
    int set_ip_tos(const int& t);
	 /** Access method.
	  * @doc get current value of IP_TOS for this socket.
	  * @return current value of IP_TOP option.
	  */
    int get_ip_tos() const;
        /** Access method.
	  * @doc Determine if socket is ready for reading/writing/exception.
	  * @param rwe Read/Write/Exception checking values correspond to:
	  * Socket::SEL_READ, SEL_WRITE, SEL_EXCEPT.
	  * @return true: if socket ready, false if not.
	  */
    virtual int select(const int rwe, double to);
	 /** Access method.
	  * @doc shutdown connection as per standard socket library.
	  * @param how method to use to shutdown socket: SHUT_RD, SHUT_WR, 
	  * SHUT_RDWR.
	  */
    virtual int shutdown(const int& how)=0;
	 /** Access method.
	  * @doc close socket connection.
	  * @return 0: on success, -1: on failure.
	  */
    virtual int close()=0;
    virtual void Print(FILE* fp=stderr);
    virtual void Error(FILE* fp=stderr);
    void Set_timeout(double sec){_timeout = sec;}
};

inline 
void Socket::Error(FILE* fp){
  switch (errno){
  case  EAGAIN :  
    fprintf(fp,
	    "The socket is marked non-blocking and no connections are present to be  accepted.");
    break;
  case  EBADF :
    fprintf(fp,"The descriptor is invalid.\n");
    break;
  case  ENOTSOCK :
    fprintf(fp,"The descriptor references a file, not a socket.\n");
    break; 
  case  EOPNOTSUPP:
    fprintf(fp,"The referenced socket is not of type SOCK_STREAM.\n");
    break;
  case  EINTR :
    fprintf(fp,"The  system  call was interrupted by a signal that was caught before a valid connection arrived.\n");
    break; 
  case  ECONNABORTED:
    fprintf(fp,"A connection has been aborted.\n");
    break; 
  case  EINVAL:
    fprintf(fp,"Socket is not listening for connections.\n");
    break; 
  case  EMFILE :
    fprintf(fp,"The per-process limit of open file descriptors has been reached.\n");
    break; 
  case  ENFILE:
    fprintf(fp,"The system maximum for file descriptors has been reached.\n");
    break; 
  case  EFAULT:
    fprintf(fp,"The addr parameter is not in a writable part of the user address space.\n");
    break; 
  case  ENOBUFS :
  case  ENOMEM :
    fprintf(fp,"Not enough free memory.  This often means that the memory allocation  is  limited by the socket buffer limits, not by the system memory.\n");
    break; 
  case  EPROTO:
    fprintf(fp," Protocol error.\n");
    break; 
  case  EPERM:
    fprintf(fp,"Firewall rules forbid connection.\n");
    break;
  }
  
}

inline
void Socket::Print(FILE* fp){
  struct sockaddr_in adr;
  socklen_t  len=sizeof(adr);
  //  char dmy[80];
  u_int16_t port;

  if( getsockname(_sockd,(sockaddr*)&adr,&len) ==0){
    port = ntohs(adr.sin_port);
    fprintf(fp,"Local ip= %s, port=%d\n", 
	    inet_ntoa(adr.sin_addr), port);
  }
  len=sizeof(adr);
  if( getpeername(_sockd,(sockaddr*)&adr,&len) ==0){
    port = ntohs(adr.sin_port);
    fprintf(fp,"Remote ip= %s, port=%d\n", 
	    inet_ntoa(adr.sin_addr), port);
  }else{
    perror("getpeername failed");
  }

  fprintf(fp,"Socket=%d, mtu=%d\n", _sockd, _mtu);
  fprintf(fp,"Sent=%Ld, Recvd=%Ld, send_call=%Ld recv_call=%Ld\n",
	  _bytes_sent, _bytes_rcvd, _send_calls, _recv_calls);
}

inline
int Socket::select(const int rwe, double to)
{
    fd_set rset, wset, eset;    
    FD_ZERO(&rset);
    wset=eset=rset;
    if (rwe & SEL_READ)
      FD_SET(_sockd, &rset);
    if (rwe & SEL_WRITE)
      FD_SET(_sockd, &wset);
    if (rwe & SEL_EXCEPT)
      FD_SET(_sockd, &eset);
    struct timeval lto;
    int sec=(int)to;
    int usec=int((to - sec)*1000000.0);
    lto.tv_sec  =sec;
    lto.tv_usec =usec;
    if (::select(_sockd+1, &rset, &wset, &eset, &lto)==-1) {
      syslog(LOG_ERR, "TCPSocket: select(): error: %s", strerror(errno));
      return(-1);
    }
    int ret=0;
    if (rwe & SEL_READ)
      ret=FD_ISSET(_sockd, &rset);
    if (rwe & SEL_WRITE)
      ret|=FD_ISSET(_sockd, &wset)<<1;
    if (rwe & SEL_EXCEPT)
      ret|=FD_ISSET(_sockd, &eset)<<2;
    return(ret);
}


#endif


