/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2004 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */


#include <TCPSocket.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <iostream>
#include <errno.h>
#include <syslog.h>

#include <Logsystem.hpp>


TCPSocket::TCPSocket(const int m):  Socket(m)
{
  _sockd=socket(PF_INET, SOCK_STREAM, 0);
  if (_sockd<=0) 
    syslog(LOG_ERR, "TCPSocket::TCPSocket(): unable to create socket.");
  set_so_reuseaddr(1);
}

TCPSocket::TCPSocket(const int& s, const int m):
		  Socket(s, m)
{
  set_so_reuseaddr(1);
}

TCPSocket::TCPSocket(const TCPSocket &s):
		  Socket(s)
{
  set_so_reuseaddr(1);
}

/** Destructor */
TCPSocket::~TCPSocket()
{
  // This is an important feature of the TCPSocket class. Especially
  // for copying of TCPSockets. TCPSockets do *not* close their
  // sockets when they are destroyed.
  syslog(LOG_ERR, "TCPSocket::~TCPSocket(): _bytes_sent=%lld", _bytes_sent);
  syslog(LOG_ERR, "TCPSocket::~TCPSocket(): _bytes_rcvd=%lld", _bytes_rcvd);
  syslog(LOG_ERR, "TCPSocket::~TCPSocket(): _send_calls=%lld", _send_calls);
  syslog(LOG_ERR, "TCPSocket::~TCPSocket(): _recv_calls=%lld", _recv_calls);
  _age.stop();
  double duration=_age.elapsed();
  double rcv_rate=_bytes_rcvd/duration*8.0/1000000;
  double send_rate=_bytes_sent/duration*8.0/1000000;
  syslog(LOG_ERR, "TCPSocket::~TCPSocket(): duration=%f", duration);
  syslog(LOG_ERR, "TCPSocket::~TCPSocket(): rcv_rate=%f", rcv_rate);
  syslog(LOG_ERR, "TCPSocket::~TCPSocket(): send_rate=%f", send_rate);

  if(Logsystem::Test_Logsys( Logsystem::INFO, 5 )){
    Logsystem::info_logger()->Tstamp()
      << "TCPSocket::~TCPSocket(): _bytes_sent=" << _bytes_sent  
      << "\n\tTCPSocket::~TCPSocket(): _bytes_rcvd=" <<  _bytes_rcvd 
      << "\n\tTCPSocket::~TCPSocket(): _send_calls=" << _send_calls  
      << "\n\tTCPSocket::~TCPSocket(): _recv_calls=" << _recv_calls  << "\n";
    Logsystem::info_logger()->Tstamp()
      << "TCPSocket::~TCPSocket(): duration=" <<  duration << " sec.\n"
      << "\tTCPSocket::~TCPSocket(): rcv_rate=" << rcv_rate << " Mbps \n"
      << "\tTCPSocket::~TCPSocket(): send_rate=" << send_rate << "Mbps. " << "\n";
  }

  if (_sockd!=-1){
    int ret = ::close(_sockd);
    if(ret !=0){
      Logsystem::error_logger()->Tstamp() << "Error in closing socket:";
      switch(ret){
      case EBADF:
	Logsystem::error_logger()->Tstamp()
	  <<" fd is not valid open file descriptor\n";
	break;
      case EINTR: 
	Logsystem::error_logger()->Tstamp()
	  <<" The close() call was interrupted by a signal.\n"; 
	break;
      case EIO: 
	Logsystem::error_logger()->Tstamp() 
	  << " An I/O error occurred.\n"; 
	break;
      default:
	Logsystem::error_logger()->Tstamp()
	  << " Undefined Error.\n"; 
	break;
      }
    }
    _sockd=-1;
  }

}

TCPSocket& TCPSocket::operator=(const TCPSocket& s)
{
    _sockd=s._sockd;
	 _mtu=s._mtu;
	 set_so_reuseaddr(1);
    return(*this);
}

void TCPSocket::set_sockd(const int& s)
{
    if (_sockd!=-1)
        ::close(_sockd);
    _sockd=s;
}

// void TCPSocket::set_mtu(const int& m){
//   _mtu=m;
// }
// retrun: 0: Succcess, -1: Failed 
int TCPSocket::bind(const string& ip, const int& port)
{
  if (ip.length()<0) {
    syslog(LOG_ERR, "TCPSocket::bind(): zero length IP address  string.");
    return -1;
  }
  struct sockaddr_in addr;
  addr.sin_family=AF_INET;
  addr.sin_port=htons((uint16_t)port);
  struct hostent *hostinfo=  gethostbyname(ip.c_str());

  if(ip =="0.0.0.0" || ip.length()==0 || hostinfo==0){ 
    addr.sin_addr.s_addr = INADDR_ANY ;
  }else{
    //      syslog(LOG_ERR, "RCPSocket::bind(): Could not get hostinfo.");
    addr.sin_addr.s_addr = *((unsigned long *)hostinfo->h_addr_list[0]);
  }
  
#ifndef LINUX
  //  addr.sin_len=sizeof(addr);
#endif // LINUX

    if (_sockd==-1) {
      syslog(LOG_ERR, "TCPSocket::bind(): Invalid sock descriptor.");
    }
    int ret=::bind(_sockd, (struct sockaddr*)&addr, sizeof(addr));
    if (ret==-1)
      syslog(LOG_ERR, "TCPSocket::bind() error: %s", strerror(errno));
    return(ret);
}

// 0: Success, -1:Failed 
int TCPSocket::listen()
{
  if (_sockd==-1)
    syslog(LOG_WARNING, "TCPSocket: Invalid sock descriptor");
  int ret=::listen(_sockd, 10);
  if (ret==-1)
    syslog(LOG_ERR, "TCPSocket: Listen error.");
  return(ret);
}

int TCPSocket::accept()
{
    if (_sockd==-1) {
        syslog(LOG_ERR, "TCPSocket::accept(): Invalid sock descriptor.");
    }
    struct sockaddr addr;
    socklen_t addrlen=sizeof(addr);
    int ret=::accept(_sockd, &addr, &addrlen);
    if (ret==-1)
      syslog(LOG_ERR, "TCPSocket: Accept error: %s", strerror(errno));
    return(ret);
}

int TCPSocket::connect(const string& ip, const int& port)
{

  int ret=0;
  if (ip.size()<=0) {
    syslog(LOG_ERR, "TCPSocket::connect(): Zero length IP address string.");
    }
  struct hostent *hostinfo =  gethostbyname(ip.c_str());
  if( hostinfo == 0){
    syslog(LOG_ERR, "TCPSocket::bind(): Could not get hostinfo.");
    return -1;
  }
  //  ret=::inet_aton((char*)ip.data(), &addr.sin_addr);
  //if (ret==0) {
  //     syslog(LOG_ERR, "Invalid IP address string");
  //   }
  struct sockaddr_in addr;


#ifndef LINUX
  //  addr.sin_len=sizeof(addr);
#endif // LINUX
  addr.sin_family=AF_INET;
  addr.sin_addr.s_addr = *((unsigned long *)hostinfo->h_addr_list[0]);
  addr.sin_port=htons((uint16_t)port);

  ret=::connect(_sockd, (struct sockaddr*)&addr, sizeof(addr));
  if (ret==-1)
    syslog(LOG_ERR, "TCPSocket: connect() error: %s", strerror(errno));

  return(ret);
}

int TCPSocket::close()
{
    return(::close(_sockd));
}

int TCPSocket::shutdown(const int& howto)
{
    if (_sockd==-1) {
        syslog(LOG_ERR, "TCPSocket::shutdown(): Attempt to close unallocated socket");
    }
    return(::shutdown(_sockd, howto));
}

ostream& operator<<(ostream& os, const TCPSocket& s)
{
  os << "TCPSocket dump\n";
  os << "    _sockd:         " << s.get_sockd() << endl;
  os << "    _bytes_sent:    " << s.get_bytes_sent() << endl;
  os << "    _bytes_rcvd:    " << s.get_bytes_rcvd() << endl;
  os << "    _send_calls:    " << s.get_send_calls() << endl;
  os << "    _recv_calls:    " << s.get_recv_calls() << endl;
  os << "-TCPSocket dump\n";
  os << "TCPSocket dump\n";
  return(os);
}

