/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2003 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */

#ifndef TCP_SOCKET_H
#define TCP_SOCKET_H

#include <common.h>
#include <Socket.h>

#include <string>
#include <iostream>
#include <syslog.h>

extern "C"{
  #include <unistd.h>
}

/** TCPSocket class
 * @doc This class encapsulates a TCP stream.
 * @auther D.Lapsley
 * Modified by M.Sekido
 */

/** TCPSocket class
 * @doc This class encapsulates a TCP Stream. It inherits from the Socket
 * class and provides methods that closely match those of the BSD sockets
 * API. The main aim of this class is to simplify the creation, 
 * connection establishment and transmission/reception procedures of TCP 
 * sockets. This class has been deliberately designed to NOT handle errors 
 * for the calling context. Instead, it returns error codes back to the 
 * calling context so that it may work out how best to deal with it. The 
 * reason for doing this is that the TCPSocket does not have the wider 
 * context of the caller and so cannot determine if an error return is 
 * indeed an error or expected behavior.
 */
class TCPSocket: public Socket 
{
  public:
	 /** Default constructor.
	  * This initializes the data structure, but does *not* create a socket.
	  */
  TCPSocket():Socket() {;}
	 /** Constructor.
	  * This constructor creates a socket and sets the internal MTU member
	  * to be equal to the supplied parameter.
	  * @param m MTU for this socket.
	  */
    TCPSocket(const int m);
	 /** Constructor.
	  * This constructor is provided two initizialization parameters.
	  * @param s socket descriptor to encapsulate.
	  * @param m MTU for this socket.
	  */
    TCPSocket(const int& s, const int m);
	 /** Copy constructor.
	  * Copies values of other TCPSocket into this object.
	  * @param s TCPSocket to copy.
	  */
    TCPSocket(const TCPSocket &s);

	 /** Destructor. */
    virtual ~TCPSocket();
	 /** Assigment operator.
	  * Copies supplied socket fields into this object.
	  * @param s socket to copy.
	  */
    TCPSocket& operator=(const TCPSocket& s);
	 /** Access method.
	  * Allows setting of TCPSocket's internal socket descriptor. This is
	  * useful for allowing a TCPSocket object to encapsulate a socket
	  * descriptor that has been returned from another function, e.g.
	  * accept().
	  * @param s socket descriptor to encapsulate.
	  */
    void set_sockd(const int& s);
	 /** Access method.
	  * Allows seting of TCPSocket's internal mtu size.
	  * @param m mtu size.
	  */
    void set_mtu(const int& m){  _mtu=m;    }
    /** Socket API method.
	  * Binds socket to supplied address.
	  * @param ip IP address to bind to.
	  * @param port port to bind to.
	  * @return 0: on success, -1 on failure.
	  */
    int bind(const string& ip, const int& port);
	 /** Socket API method.
	  * Prepare a socket for accepting incoming connections.
	  * @return 0: on success, -1 on failure.
	  */
    int listen();
	 /** Socket API method.
	  * Wait for incoming connections on a socket. 
	  * @return -1: on failure, positive integer on success. On succes
	  *         the return value is a valid socket descriptor.
	  */
    int accept();
	 /** Socket API method.
	  * Attempt to establish a connection to the supplied address.
	  * @param ip IP address to connect to.
	  * @param port port to connect to.
	  * @return 0: on success, -1 on failure.
	  */
    int connect(const string& ip, const int& port);
	 /** Socket API method.
	  * Attempt to receive data into supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes read.
	  */
    int recv(u_int8_t* b, const int& n);
	 /** Socket API method.
	  * Attempt to send data in supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes sent.
	  */
    int readline(uint8_t* b, const int n);
    int send(u_int8_t* b, const int& n);
    int sendto(const struct sockaddr* to, const int tolen, char* s, 
	       const int n){
      cerr << "Warning TCPSocket::sendto(): Do Nothing in TCP" << endl; 
      return -1;
    }
    int recvfrom(struct sockaddr& from, int& fromlen, char* s, const int& n){
      cerr << "Warning TCPSocket::recvfrom(): Do Nothing in TCP" << endl; 
      return -1;
    }
    /** Socket API method.
	  * @doc shutdown connection as per standard socket library.
	  * @param how method to use to shutdown socket: SHUT_RD, SHUT_WR, 
	  * SHUT_RDWR.
	  */
    int shutdown(const int& howto);
	 /** Socket API method.
	  * @doc close socket connection.
	  */
    int close();
};

ostream& operator<<(ostream& os, const TCPSocket& s);

inline 
int TCPSocket::recv(u_int8_t* b, const int& n)
{
  int length=0;
  int iter=0;
  int ret=0;

  
#if 1
  fd_set rfd;
  struct timeval timeout;
  timeout.tv_sec=IO_TIMEOUT;
  timeout.tv_usec=0;
  int nd=_sockd+1;
  do {
    FD_SET(_sockd, &rfd);
    ret=::select(nd, &rfd, NULL, NULL, &timeout);
    if (ret==0) {
      syslog(LOG_ERR, "TCPSocket::recv(): timeout expired");
      fprintf(stderr, "TCPSocket::recv(): timeout expired");
      return(-1);
    }
    if (ret<0) {
      syslog(LOG_ERR, "TCPSocket::recv(): select error: %s",
	     strerror(errno));
      fprintf(stderr, "TCPSocket::recv(): select error: %s",
	      strerror(errno));
    }
    if (FD_ISSET(_sockd, &rfd)) {
      ret=read(_sockd, b, n-length);
      //      ret=::recv(_sockd, b, n-length,flag);
      ++_recv_calls;
      if (ret>0) {
	_bytes_rcvd+=ret;
	length+=ret;
	if (length<n) {
	  b+=ret;
	} else if (length>=n) {
	  return(length);
	}
      } else if (ret<0 && errno!=EINTR) {
	syslog(LOG_ERR, "TCPSocket::recv(): %s", strerror(errno));
	fprintf(stderr, "TCPSocket::recv(): %s", strerror(errno));
	return(-1);
      }
      //      else if (ret ==0) return length;
      
    }
    if (++iter>10) {
      syslog(LOG_ERR, "TCPSocket::recv(): exceeded max iter");
      return(length==0?-1:length);
    }
  } while ( (length<n) || (ret<0 && errno==EINTR) );
#else
  u_int8_t* b0 =b;  
  do {
    ret=::read(_sockd, b, n-length); 
    //    ret=::recv(_sockd, b, n-length,flag); 
    b[ret]='\0';    fprintf(stderr,"read retrun =%d, %s\n",ret,b);
    ++_recv_calls;
    if (ret>0) {
      _bytes_rcvd+=ret;
      length+=ret;
      if (length<n) {
	b+=ret;
      } else if (length>=n) {
	b0[length]='\0';
	fprintf(stderr,"Received data=%s\n",b0);
	return(length);
      }
    } else if (ret<0 && errno!=EINTR) {
      syslog(LOG_ERR, "TCPSocket::recv(): %s", strerror(errno));
      fprintf(stderr, "TCPSocket::recv(): %s", strerror(errno));
      return(-1);
    } else if (ret==0) {
      syslog(LOG_ERR, "TCPSocket::recv(): zero byte read");
      return(length);
    }
    if (++iter>10) {
      syslog(LOG_ERR, "TCPSocket::recv(): exceeded max iter");
      return(length==0?-1:length);
    }
  } while ( (length<n) || (ret<0 && errno==EINTR) );
#endif
	 
  syslog(LOG_ERR, "TCPSocket::recv(): %s", strerror(errno));
  return(-1);
}

// Return: >0 = Success, 0=Closed, -1=Failed
inline 
int TCPSocket::send( u_int8_t* b, const int& n)
{
  int sent=0;
  int iter=0;
  int ret=0;
#if 0
  fd_set wfd;
  struct timeval timeout;
  timeout.tv_sec=IO_TIMEOUT;
  timeout.tv_usec=0;
  int nd=_sockd+1;
  do {
    FD_SET(_sockd, &wfd);
    ret=::select(nd, NULL, &wfd, NULL, &timeout);
    fprintf(stderr,"select ret= %d, ", ret);
    if (ret==0) {
      syslog(LOG_ERR, "TCPSocket::send(): select timeout");
      return(-1);
    }
    if (ret<0) {
      syslog(LOG_ERR, "TCPSocket::send(): select error: %s",
	     strerror(errno));
      return(-1);
    }
    if (FD_ISSET(_sockd, &wfd)) {
      //      ret=::write(_sockd, b, n-sent);
      ret=::send(_sockd, b, n-sent, MSG_NOSIGNAL);
      fprintf(stderr,"write ret= %d\n", ret);
      if (ret>0) {
	_bytes_sent+=ret;
	sent+=ret;
	if (sent<n) {
	  b+=ret;
	} else if (sent>=n) {
	  return(sent);
	}
      } else if (ret<0 && errno!=EINTR) {
	fprintf(stderr,"TCPSocket::send(): %s", strerror(errno));
	syslog(LOG_ERR, "TCPSocket::send(): %s", strerror(errno));
	return(-1);
      }
//       else{
// 	// Inserted by M.Sekido 2007.1.5
// 	fprintf(stderr,"TCPSocket::send():EINTR %s", strerror(errno));
// 	return(-1);
//       }
    }
    if (++iter>10) {
      syslog(LOG_ERR, "TCPSocket::send(): exceeded max iter");
      return(sent==0?-1:sent);
    }
  } while ( (sent<n) || (ret<0 && errno==EINTR) );						 
#else
  do {
    //    ret=::write(_sockd, b, n-sent);
    ret=::send(_sockd, b, n-sent,MSG_NOSIGNAL);
    //    fprintf(stderr,"write ret=%d\n", ret);
    ++_send_calls;
    if (ret>0) {
      sent+=ret;
      _bytes_sent+=ret;
      if (sent<n) {
	b+=ret;
      } else if (sent>=n) {
	return(sent);
      }
    } else if (ret<0 && errno!=EINTR) {
      //    } else if (ret<0 && ) {
      syslog(LOG_ERR, "TCPSocket::send(): %s", strerror(errno));
      fprintf(stderr, "TCPSocket::send(): %s", strerror(errno));
      Socket::Error();
      return(-1);
    } else if (ret==0) {
      syslog(LOG_ERR, "TCPSocket::send(): zero byte send()");
      return(sent);
    }
    
    if (++iter>10) {
      syslog(LOG_ERR, "TCPSocket::send(): exceeded max iter");
      return(sent==0?-1:sent);
    }
    //  } while ( (sent<n) );
  } while ( (sent<n) || (ret<0 && errno==EINTR) );						  
#endif
  syslog(LOG_ERR, "TCPSocket::send(): %s", strerror(errno));
  return(-1);
}

inline 
int TCPSocket::readline(uint8_t* b, const int n)
{
  int length=0;
  int iter=0;
  int ret=0;
  int flag=0;
  //  const int bufsize=256;
  //  uint8_t readbuf[bufsize](0);

  do {
    //    ret=::read(_sockd, readbuf, bufsize); 
    ret=::recv(_sockd, b+length, n-length,flag); 
    ++_recv_calls;

    if (ret>0) {
      _bytes_rcvd+=ret;
      length+=ret;
      b[ret]='\0';
      fprintf(stderr,"Received data=%s",b);
      //      if(readbuf[ret-1] == '\n' || length >= n ){
      if(b[length+ret-1] == '\n' || length >= n ){
	fprintf(stderr,"Received data=%s",b);
	return(length);
      }
      return(length);
    } else if (ret<0 && errno!=EINTR) {
      syslog(LOG_ERR, "TCPSocket::recv(): %s", strerror(errno));
      fprintf(stderr, "TCPSocket::recv(): %s", strerror(errno));
      return(-1);
    } else if (ret==0) {
      syslog(LOG_ERR, "TCPSocket::recv(): zero byte read");
      return(length);
    }

    if (++iter>10) {
      syslog(LOG_ERR, "TCPSocket::recv(): exceeded max iter");
      return(length==0?-1:length);
    }
  } while ( (length<n) || (ret<0 && errno==EINTR) );
	 
  syslog(LOG_ERR, "TCPSocket::recv(): %s", strerror(errno));
  return(-1);
}



#endif

