/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2004 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <iostream>
#include <syslog.h>
#include "UDPSocket.h"

UDPSocket::UDPSocket():
  Socket()
{
  // NO socket created.
}

UDPSocket::UDPSocket(const int m):
  Socket(m)
{
  //  _sockd=socket(PF_INET, SOCK_DGRAM, 0);
  _sockd=socket(AF_INET,SOCK_DGRAM, IPPROTO_UDP); 

  if (_sockd<=0) 
    syslog(LOG_ERR, "UDPSocket::UDPSocket(): unable to create socket.");
  set_so_reuseaddr(1);
}

UDPSocket::UDPSocket(const int& s, const int m):
  Socket(s, m)
{
	 set_so_reuseaddr(1);
}

UDPSocket::UDPSocket(const UDPSocket& s):
  Socket(s)
{
	 set_so_reuseaddr(1);
}

UDPSocket::~UDPSocket()
{
  if (_sockd!=-1)
    syslog(LOG_ERR, "UDPSocket::~UDPSocket(): _bytes_sent==%lld", _bytes_sent);
  syslog(LOG_ERR, "UDPSocket::~UDPSocket(): _bytes_rcvd==%lld", _bytes_rcvd);
  syslog(LOG_ERR, "UDPSocket::~UDPSocket(): _send_calls==%lld", _send_calls);
  syslog(LOG_ERR, "UDPSocket::~UDPSocket(): _recv_calls==%lld", _recv_calls);
  
  
  _age.stop();
  double duration=_age.elapsed();
  double rcv_rate=_bytes_rcvd/duration*8.0/1000000;
  double send_rate=_bytes_sent/duration*8.0/1000000;
  syslog(LOG_ERR, "UDPSocket::~UDPSocket(): duration==%f", duration);
  syslog(LOG_ERR, "UDPSocket::~UDPSocket(): rcv_rate==%f", rcv_rate);
  syslog(LOG_ERR, "UDPSocket::~UDPSocket(): send_rate==%f", send_rate);

  if(Logsystem::Test_Logsys( Logsystem::INFO, 5 )){
    Logsystem::info_logger()->Tstamp()
      <<  "UDPSocket::~UDPSocket(): _bytes_sent=" <<  _bytes_sent 
      <<  "\n\tUDPSocket::~UDPSocket(): _bytes_rcvd=" << _bytes_rcvd 
      <<  "\n\tUDPSocket::~UDPSocket(): _send_calls=" << _send_calls
      <<  "\n\tUDPSocket::~UDPSocket(): _recv_calls=" << _recv_calls << "\n";
    Logsystem::info_logger()->Tstamp()
      << "UDPSocket::~UDPSocket(): duration=" << duration << "sec.\n"
      << "\tUDPSocket::~UDPSocket(): rcv_rate=" << rcv_rate << " Mbps\n"
      << "\tUDPSocket::~UDPSocket(): send_rate=" << send_rate << "\n";
  }

  if (_sockd!=-1){
    int ret = ::close(_sockd);
    if(ret !=0){
      cerr << "Error in closing socket:";
      switch(ret){
      case EBADF:
	cerr << " fd is not valid open file descriptor\n"; break;
      case EINTR: 
	cerr << " The close() call was interrupted by a signal.\n"; break;
      case EIO: 
	cerr << " An I/O error occurred.\n"; break;
      default:
	//	       cerr << " Undefined Error.\n";  break;
	perror( "");  break;
      }
    }
    _sockd=-1;
  }
}

UDPSocket& UDPSocket::operator=(const UDPSocket& s)
{
    _sockd=s._sockd;
	 _mtu=s._mtu;
	 set_so_reuseaddr(1);
    return(*this);
}

void UDPSocket::set_sockd(const int& s)
{
    if (_sockd!=-1)
        ::close(_sockd);
    _sockd=s;
}

int UDPSocket::bind(const string& ip, const int& port)
{
  //    const char* const p=(char*)ip.c_str();
    struct sockaddr_in addr;
    //    int ret=0;
    //     if (ip.size()<0) {
    //         syslog(LOG_ERR, "UDPSocket::bind(): Zero length IP address string.");
    //     }      
    //    ret=::inet_aton(p, &addr.sin_addr);
  
#ifndef LINUX
    //    addr.sin_len=sizeof(addr);
#endif // LINUX
    addr.sin_family=(uint16_t)AF_INET;
    addr.sin_port=htons((uint16_t)port);
    if(ip =="0.0.0.0" || ip.length()==0){ 
      addr.sin_addr.s_addr = INADDR_ANY ;
    }else{
      //      syslog(LOG_ERR, "RCPSocket::bind(): Could not get hostinfo.");
      struct hostent *hostinfo =  gethostbyname(ip.c_str());
      addr.sin_addr.s_addr = *((unsigned long *)hostinfo->h_addr_list[0]);
    }

    int ret=::bind(_sockd, (struct sockaddr*)&addr, sizeof(addr));
    if (_sockd==-1) {
      Logsystem::error_logger()->Tstamp() 
	<<"UDPSocket::bind(): Invalid sock descriptor." << strerror(errno) << "\n";
      syslog(LOG_ERR, "UDPSocket::bind(): Invalid sock descriptor.");
    }
    if (ret==-1) {
      Logsystem::error_logger()->Tstamp() 
	<< "UDPSocket::bind(): error " << strerror(errno) << "\n";
      syslog(LOG_ERR, "UDPSocket::bind(): error %s", strerror(errno));
    }
    return(ret);

}

int UDPSocket::listen()
{
	 syslog(LOG_WARNING, "UDPSocket::listen(): call has no effect.");
	 return(0);
}

int UDPSocket::accept()
{
	 syslog(LOG_WARNING, "UDPSocket::accept(): call has no effect.");
	 return(0);
}

int UDPSocket::connect(const string& ip, const int& port)
{

  if (ip.size()<=0) {
    syslog(LOG_ERR, "UDPSocket::connectZero length IP string.");
  }
  fprintf(stderr,"Calling UDPScoket::connect(%s, %d) \n", ip.c_str(), port);

  struct hostent *hostinfo =  gethostbyname(ip.c_str());
  if( hostinfo == NULL){
    syslog(LOG_ERR, "UDPSocket::bind(): Could not get hostinfo.");
    return -1;
  }

  struct sockaddr_in addr;
  int ret=0;
  
//     ret=::inet_aton((char*)ip.data(), &addr.sin_addr);
//     if (ret==0) {
//         syslog(LOG_ERR, "UDPSocket::connect(): Invalid IP address string: %s", strerror(errno));
//     }

    addr.sin_port=htons((uint16_t)port);
#ifndef LINUX
    //    addr.sin_len=sizeof(addr);
#endif // LINUX
    addr.sin_family=AF_INET;
    addr.sin_addr.s_addr = *((unsigned long *)hostinfo->h_addr_list[0]);
  
    ret=::connect(_sockd, (struct sockaddr*)&addr, sizeof(addr));
    if (ret==-1) {
      fprintf(stderr, "UDPSocket::connect(): error: %s", strerror(errno));
      syslog(LOG_ERR, "UDPSocket::connect(): error: %s", strerror(errno));
    }
    return(ret);
}


int UDPSocket::close()
{
    return(::close(_sockd));
}

int UDPSocket::shutdown(const int& howto)
{
    if (_sockd==-1) {
        syslog(LOG_ERR, "UDPSocket::shutdown(): Attempt to close unallocated socket");
    }
    return(::shutdown(_sockd, howto));
}

ostream& operator<<(ostream& os, const UDPSocket& s)
{
    os << "UDPSocket dump\n";
    os << "    _sockd:         " << s.get_sockd() << endl;
    os << "    _bytes_sent:    " << s.get_bytes_sent() << endl;
    os << "    _bytes_rcvd:    " << s.get_bytes_rcvd() << endl;
    os << "    _send_calls:    " << s.get_send_calls() << endl;
    os << "    _recv_calls:    " << s.get_recv_calls() << endl;
    os << "-UDPSocket dump\n";
    return(os);
}

#ifndef _SELECT_
int UDPSocket::recvfrom(struct sockaddr& from, int& fromlen,
			char* s, const int& n)
{
  //  int rdflag = MSG_DONTWAIT;
  int ret=::recvfrom(_sockd, (void*)&s[0], n, 0, 
		     (struct sockaddr*)&from, (socklen_t*)&fromlen);
  _bytes_rcvd += ret;
  return(ret);
}

#endif
