
/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2004 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */


/** UDPSocket class.
 * @doc This class encapsulates a UDP packet stream. 
 * @author David Lapsley (dlapsley@haystack.mit.edu)
 */ 
/** 
 */

#ifndef UDP_SOCKET_H
#define UDP_SOCKET_H

#include <common.h>
#include <Socket.h>
#include <syslog.h>
#include "Logsystem.hpp"

extern "C"{
  #include <unistd.h>
}

/** UDPSocket class.
 * @doc This class encapsulates a UDP packet stream. It inherits from the 
 * socket class and provides methods that closely match those of the BSD 
 * sockets API. The main aim of this class is to simplify the creation and
 * data transmission/reception procedures of UDP sockets.  This class has been   
 * deliberately designed to NOT handle errors for the calling context. 
 * Instead, it returns error codes back to the calling context so that it may 
 * work out how best to deal with it. The reason for doing this is that the 
 * TCPSocket does not have the wider context of the caller and so cannot 
 * determine if an error return is indeed an error or expected behavior.
 */
class UDPSocket: public Socket 
{
 public:
    /** Default constructor.
	  * This calls the base constructor to create an opened UDP
	  * socket (i.e. socket of type SOCK_DGRAM)
	  */
    UDPSocket();
	 /** Constructor/
	  * This constructo has a single initialization parameter.
	  * @param m MTU for this socket.
	  */
    UDPSocket(const int m);
        /** Constructor.
	  * This constructor is provided two initizialization parameters.
	  * @param s socket descriptor to encapsulate.
	  * @param m MTU for this socket.
	  */
    UDPSocket(const int& s, const int m);
        /** Copy constructor.
	  * Copies values of other UDPSocket into this object.
	  * @param s UDPSocket to copy.
	  */
    UDPSocket(const UDPSocket& s);
         /** Destructor. */
    virtual ~UDPSocket();
         /** Assigment operator.
	  * Copies supplied socket fields into this object.
	  * @param s socket to copy.
	  */
    UDPSocket& operator=(const UDPSocket& s);
        /** Access method.
	  * Allows setting of TCPSocket's internal socket descriptor. This is
	  * useful for allowing a TCPSocket object to encapsulate a socket
	  * descriptor that has been returned from another function, e.g.
	  * accept().
	  * @param s socket descriptor to encapsulate.
	  */
    void set_sockd(const int& s);
       /** Socket API method.
	  * Binds socket to supplied address.
	  * @param ip IP address to bind to.
	  * @param port port to bind to.
	  * @return 0: on success, -1 on failure.
	  */
    int bind(const string& ip, const int& port);
	 /** Socket API method.
	  * Prepare a socket for accepting incoming connections.
	  * @return 0: on success, -1 on failure.
	  */
    int listen();
	 /** Socket API method.
	  * Wait for incoming connections on a socket. 
	  * @return -1: on failure, positive integer on success. On succes
	  *         the return value is a valid socket descriptor.
	  */
    int accept();
       /** Socket API method.
	  * Attempt to establish a connection to the supplied address.
	  * Note that UDP is connectionless and that in this case "connection" 
	  * refers to the fact that the destination address fields are 
	  * set to the parameters supplied in order that subsequent calls to
	  * send() use these parameters. Otherwise, an address must be supplied
	  * with each call to sendto().
	  * @param ip IP address to connect to.
	  * @param port port to connect to.
	  * @return 0: on success, -1 on failure.
	  */
    int connect(const string& ip, const int& port);
	 /** Socket API method.
	  * Attempt to receive data into supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes read.
	  */
    int recv(u_int8_t* b, const int& n);
	 /** Socket API method.
	  * Attempt to send data in supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes sent.
	  */
    int send(u_int8_t* b, const int& n);
    /** Socket API method.
	  * Attempt to recv data from supplied buffer and store the destination.
	  * @param from Place to store address of sender.
	  * @param fromlen Length of from address.
	  * @param s socket buffer to store data in.
	  * @param n number of bytes to attempt to recv.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes received.
	  */
    int recvfrom(struct sockaddr& from, int& fromlen, char* s, 
		 const int& n);
    /** Socket API method.
	  * Attempt to send data from supplied buffer to specified destination.
	  * @param ip IP address to send data to.
	  * @param port port to send data to.
	  * @param s socket buffer to from which to get data to send.
	  * @param n number of bytes to attempt to send.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes sent.
	  */
    int sendto(const struct sockaddr* to, const int tolen, 
					char* s, const int n);
    /** Socket API method.
	  * @doc shutdown connection as per standard socket library.
	  * @param how method to use to shutdown socket: SHUT_RD, SHUT_WR, 
	  * SHUT_RDWR.
	  */
    int shutdown(const int& howto);
    /** Socket API method.
	  * @doc close socket connection.
	  * @return 0: on success, -1: on failure.
	  */
    int close();
};

ostream& operator<<(ostream& os, const UDPSocket& s);

inline int UDPSocket::recv(u_int8_t* b, const int& n)
{
  int length=0;
  int iter=0;
  int ret=0;
#if 1
  fd_set rfd;
  struct timeval timeout;
  timeout.tv_sec= (time_t)_timeout;
  timeout.tv_usec=(useconds_t)(_timeout*1000000);
  int nd=_sockd+1;
  do {
    FD_SET(_sockd, &rfd);
    ret=::select(nd, &rfd, NULL, NULL, &timeout);
    if (ret==0) {
      syslog(LOG_ERR, "UDPSocket::recv(): timeout expired");
      fprintf(stderr, "UDPSocket::recv(): timeout expired");
      return(-1);
    }
    if (ret<0) {
      syslog(LOG_ERR, "UDPSocket::recv(): select error: %s",
	     strerror(errno));
      fprintf(stderr, "UDPSocket::recv(): select error: %s",
	      strerror(errno));
    }
    if (FD_ISSET(_sockd, &rfd)) {
      ret=read(_sockd, b, n-length);
      //      ret=::recv(_sockd, b, n-length,flag);
      ++_recv_calls;
      if (ret>0) {
	_bytes_rcvd+=ret;
	length+=ret;
	if (length<n) {
	  b+=ret;
	} else if (length>=n) {
	  return(length);
	}
      } else if (ret<0 && errno!=EINTR) {
	syslog(LOG_ERR, "UDPSocket::recv(): %s", strerror(errno));
	fprintf(stderr, "UDPSocket::recv(): %s", strerror(errno));
	return(-1);
      }
      //      else if (ret ==0) return length;
      
    }
    if (++iter>10) {
      syslog(LOG_ERR, "UDPSocket::recv(): exceeded max iter");
      fprintf(stderr, "UDPSocket::recv(): exceeded max iter");
      return(length==0?-1:length);
    }
  } while ( (length<n) || (ret<0 && errno==EINTR) );

#else
  do {
    ret=std::read(_sockd, b, n-length);
    ++_recv_calls;
    if (ret>0) {
      _bytes_rcvd+=ret;
      length+=ret;
      if (length<n) {
	b+=ret;
      } else if (length>=n) {
	return(length);
      }
    } else if (ret<0 && errno!=EINTR) {
      syslog(LOG_ERR, "recv(): %s", strerror(errno));
      return(-1);
    }
    if (++iter>10) {
      syslog(LOG_ERR, "recv(): exceeded max iter");
      return(length);
    }
  } while ( (length<n) || (ret<0 && errno==EINTR) );
#endif

  syslog(LOG_ERR, "recv(): %s", strerror(errno));
  return(-1);
}

inline
int UDPSocket::send(u_int8_t* b, const int& n)
{
	 int sent=0;
	 int iter=0;
	 int ret=0;

	 do {
	   //	   ret=::write(_sockd, b, n-sent);
	   ret=::send(_sockd, b, n-sent, MSG_NOSIGNAL);
	   ++_send_calls;
	   if (ret>0) {
	     sent+=ret;
	     _bytes_sent+=ret;
	     if (sent<n) {
	       b+=ret;
	     } else if (sent>=n) {
	       return(sent);
	     }
	   } else if (ret<0 && errno!=EINTR) {
	     syslog(LOG_ERR, "send(): %s", strerror(errno));
	     fprintf(stderr, "send(): %s\n", strerror(errno));
	     return(-1);
	   }
	   if (++iter>10) {
	     syslog(LOG_ERR, "sent(): exceeded max iter");
	     return(sent);
	   }
	 } while ( (sent<n) || (ret<0 && errno==EINTR) );						  
	 syslog(LOG_ERR, "send(): %s", strerror(errno));
	 return(-1);
}

inline int UDPSocket::sendto(const struct sockaddr* to, const int tolen,
			     char* s, const int n){
  int ret=::sendto(_sockd, (const char*)&s[0], n, 0, to, tolen);
  if (ret<0){
    switch (errno){
    case EINTR: 
      //      syslog(LOG_ERR, "send(): %s", strerror(errno));
      fprintf(stderr, "Errorin send(): Signal issued before sending data.\n");
      fprintf(stderr, "\t: %s\n", strerror(errno));
      return(-1); break;
    case EBADF:
      fprintf(stderr, "Errorin send(): Bad descriptor.\n");
      fprintf(stderr, "\t: %s\n", strerror(errno));
      return(-1); break;
    case ECONNRESET:
      fprintf(stderr, "Errorin send(): Connection Rset by destination.\n");
      fprintf(stderr, "\t: %s\n", strerror(errno));
      return(-1); break;
    case EMSGSIZE:
      fprintf(stderr, "Errorin send(): Too Large message for this socket.\n");
      fprintf(stderr, "\t: %s\n", strerror(errno));
      return(-1); break;
    case ENOMEM:
      fprintf(stderr, "Errorin send(): Memory is exhousted.\n");
      fprintf(stderr, "\t: %s\n", strerror(errno));
      return(-1); break;
    case ENOTCONN:
      fprintf(stderr, "Errorin send(): Lack of connection and destination.\n");
      fprintf(stderr, "\t: %s\n", strerror(errno));
      return(-1); break;
    case ENOTSOCK:
      fprintf(stderr, "Errorin send(): Missing Socket.\n");
      fprintf(stderr, "\t: %s\n", strerror(errno));
      return(-1); break;
    case EOPNOTSUPP:
      fprintf(stderr, "Errorin send(): Illegal flag is assigned.\n");
      fprintf(stderr, "\t: %s\n", strerror(errno));
      return(-1); break;
    default:
      fprintf(stderr, "Error in send(): %s\n", strerror(errno));
      return(-1);
    }
    
  }
  _bytes_sent += ret;
  return(ret);
}

#ifdef _SELECT_ 
inline int UDPSocket::recvfrom(struct sockaddr& from, int& fromlen,
			char* s, const int& n)
{
  
  fd_set rfd;
  struct timeval timeout;
  timeout.tv_sec= (time_t)_timeout;
  timeout.tv_usec=(useconds_t)(_timeout*1000000);
  int nd=_sockd+1;
  int ret =0;
  int rdflag = MSG_DONTWAIT;

  FD_SET(_sockd, &rfd);
  int iret=::select(nd, &rfd, NULL, NULL, &timeout);
  if (iret==0) {
    Logsystem::error_logger()->Tstamp() 
      << "UDPSocket::recvfrom(): timeout expired" << endl;
    return(-1);
  }
  if (iret<0) {
    Logsystem::error_logger()->Tstamp() 
      <<  "UDPSocket::recvfrom(): select error: " 
      <<   strerror(errno) << endl;
    return -1;
  }
  
  if (FD_ISSET(_sockd, &rfd)) {
    ret=::recvfrom(_sockd, (void*)&s[0], n, rdflag, (struct sockaddr*)&from, 
		   (socklen_t*)&fromlen);
    if(ret >0)   _bytes_rcvd += ret;
  }

  
  return(ret);
}
#endif

#endif

