/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2003 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */

#ifdef UDT_SUPPORTED

#ifndef UDT2_SOCKET_H
#define UDT2_SOCKET_H

#include <common.h>
#include <Socket.h>
#include <udt.h>
#include <string>
#include <iostream>
#include <syslog.h>

using namespace UDT;

/** UDT2Socket class
 * @doc This class encapsulates a TCP stream.
 */

/** UDT2Socket class
 * @doc This class encapsulates a TCP Stream. It inherits from the Socket
 * class and provides methods that closely match those of the BSD sockets
 * API. The main aim of this class is to simplify the creation, 
 * connection establishment and transmission/reception procedures of TCP 
 * sockets. This class has been deliberately designed to NOT handle errors 
 * for the calling context. Instead, it returns error codes back to the 
 * calling context so that it may work out how best to deal with it. The 
 * reason for doing this is that the UDT2Socket does not have the wider 
 * context of the caller and so cannot determine if an error return is 
 * indeed an error or expected behavior.
 */
class UDT2Socket: public Socket 
{
  public:
	 /** Default constructor.
	  * This initializes the data structure, but does *not* create a socket.
	  */
    UDT2Socket();
	 /** Constructor.
	  * This constructor creates a socket and sets the internal MTU member
	  * to be equal to the supplied parameter.
	  * @param m MTU for this socket.
	  */
    UDT2Socket(const int &m);
	 /** Constructor.
	  * This constructor is provided two initizialization parameters.
	  * @param s socket descriptor to encapsulate.
	  * @param m MTU for this socket.
	  * @note For UDTSockets, UDTSOCKET descriptor type is an integer.
	  * So, there is no need to change method signatures. However,
	  * UDTSOCKET descriptors are handled in a separate table.
	  */
    UDT2Socket(const int& s, const int& m);
	 /** Copy constructor.
	  * Copies values of other UDT2Socket into this object.
	  * @param s UDT2Socket to copy.
	  */
    UDT2Socket(const UDT2Socket &s);
	 /** Destructor. */
    virtual ~UDT2Socket();
	 /** Assigment operator.
	  * Copies supplied socket fields into this object.
	  * @param s socket to copy.
	  */
    UDT2Socket& operator=(const UDT2Socket& s);
	 /** Access method.
	  * Allows setting of UDT2Socket's internal socket descriptor. This is
	  * useful for allowing a UDT2Socket object to encapsulate a socket
	  * descriptor that has been returned from another function, e.g.
	  * accept(). This is the default version that uses integer
	  * descriptors.
	  * @param s socket descriptor to encapsulate.
	  */
    void set_sockd(const int& s);
	 /** Access method.
	  * Allows seting of UDT2Socket's internal mtu size.
	  * @param m mtu size.
	  */
	 void set_mtu(const int& m);
	 /** Socket API method.
	  * Binds socket to supplied address.
	  * @param ip IP address to bind to.
	  * @param port port to bind to.
	  * @return 0: on success, -1 on failure.
	  */
    int bind(const string& ip, const int& port);
	 /** Socket API method.
	  * Prepare a socket for accepting incoming connections.
	  * @return 0: on success, -1 on failure.
	  */
    int listen();
	 /** Socket API method.
	  * Wait for incoming connections on a socket. 
	  * @deprecated for use with UDT2Socket. Use UDTSOCKET version.
	  * @return -1: on failure, positive integer on success. On succes
	  *         the return value is a valid socket descriptor.
	  */
    int accept();
	 /** Socket API method.
	  * Attempt to establish a connection to the supplied address.
	  * @param ip IP address to connect to.
	  * @param port port to connect to.
	  * @return 0: on success, -1 on failure.
	  */
    int connect(const string& ip, const int& port);
	 /** Socket API method.
	  * Attempt to receive data into supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes read.
	  */
    int recv(u_int8_t* b, const int& n);
	 /** Socket API method.
	  * Attempt to send data in supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes sent.
	  */
    int send(u_int8_t* b, const int& n);
    /** Socket API method.
	  * @doc shutdown connection as per standard socket library.
	  * @param how method to use to shutdown socket: SHUT_RD, SHUT_WR, 
	  * SHUT_RDWR.
	  */
    int shutdown(const int& howto);
	 /** Socket API method.
	  * @doc close socket connection.
	  */
    int close();
};

ostream& operator<<(ostream& os, const UDT2Socket& s);

inline 
int UDT2Socket::recv(u_int8_t* b, const int& n)
{
	 int length=0;
	 int iter=0;
	 int ret=0;
	 do {
		  ret=UDT::recv(_sockd, (char*)b, n-length, 0);
		  ++_recv_calls;
		  if (ret>0) {
				_bytes_rcvd+=ret;
				length+=ret;
				if (length<n) {
					 b+=ret;
				} else if (length>=n) {
					 return(length);
				}
		  } else if (ret<0 && errno!=EINTR) {
				syslog(LOG_ERR, "UDT2Socket::recv(): %s", strerror(errno));
				return(-1);
		  } else if (ret==0) {
				syslog(LOG_ERR, "UDT2Socket::recv(): zero byte read");
				return(length);
		  }
		  if (++iter>10) {
				syslog(LOG_ERR, "UDT2Socket::recv(): exceeded max iter");
				return(length==0?-1:length);
		  }
	 } while ( (length<n) || (ret<0 && errno==EINTR) );
	 syslog(LOG_ERR, "UDT2Socket::recv(): %s", strerror(errno));
	 return(-1);
}

inline 
int UDT2Socket::send(u_int8_t* b, const int& n)
{
	 int sent=0;
	 int iter=0;
	 int ret=0;
	 do {
		  ret=UDT::send(_sockd, (char*)b, n-sent, 0);
		  ++_send_calls;
		  if (ret>0) {
				sent+=ret;
				_bytes_sent+=ret;
				if (sent<n) {
					 b+=ret;
				} else if (sent>=n) {
					 return(sent);
				}
		  } else if (ret<0 && errno!=EINTR) {
				syslog(LOG_ERR, "UDT2Socket::send(): %s", strerror(errno));
				return(-1);
		  } else if (ret==0) {
				syslog(LOG_ERR, "UDT2Socket::send(): zero byte send()");
				return(sent);
		  }
		  if (++iter>10) {
				syslog(LOG_ERR, "UDT2Socket::send(): exceeded max iter");
				return(sent==0?-1:sent);
		  }
	 } while ( (sent<n) || (ret<0 && errno==EINTR) );						  
	 syslog(LOG_ERR, "UDT2Socket::send(): %s", strerror(errno));
	 return(-1);
}

#endif

#endif // UDT_SUPPORTED


