/* 
 * This software is released under the terms of the MIT License(included below). 
 *  
 * Copyright (c) 2004 MIT Haystack Observatory 
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a   
 * copy of this software and associated documentation files (the "Software"),    
 * to deal in the Software without restriction, including without limitation   
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,   
 * and/or sell copies of the Software, and to permit persons to whom the  
 * Software is furnished to do so, subject to the following conditions: 
 *  
 * The above copyright notice and this permission notice shall be included in    
 * all copies or substantial portions of the Software. 
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE  
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER   
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE. 
 * 
 */


#ifndef UNIX_SOCKET_H
#define UNIX_SOCKET_H

#include <common.h>
#include <Socket.h>
#include <list>
#include <string>
#include <iostream>

extern "C"{
  #include <unistd.h>
}

/** UnixSocket class
 * @doc This class encapsulates a TCP stream.
 */

/** UnixSocket class
 * @doc This class encapsulates a TCP Stream. It inherits from the Socket
 * class and provides methods that closely match those of the BSD sockets
 * API. The main aim of this class is to simplify the creation, 
 * connection establishment and transmission/reception procedures of TCP 
 * sockets. This class has been deliberately designed to NOT handle errors 
 * for the calling context. Instead, it returns error codes back to the 
 * calling context so that it may work out how best to deal with it. The 
 * reason for doing this is that the UnixSocket does not have the wider 
 * context of the caller and so cannot determine if an error return is 
 * indeed an error or expected behavior.
 */
class UnixSocket: public Socket 
{
  public:
	 /** Default constructor.
	  * This calls the base constructor to create an opened TCP
	  * socket (i.e. socket of type SOCK_STREAM)
	  */
    UnixSocket();
	 /** Constructor.
	  * This constructor creates a socket and sets the internal MTU member
	  * to be equal to the supplied parameter.
	  * @param m MTU for this socket.
	  */
    UnixSocket(const int &m);
	 /** Copy constructor.
	  * Copies values of other UnixSocket into this object.
	  * @param s UnixSocket to copy.
	  */
    UnixSocket(const UnixSocket &s);
	 /** Constructor.
	  * This constructor is provided two initizialization parameters.
	  * @param s socket descriptor to encapsulate.
	  * @param m MTU for this socket.
	  */
    UnixSocket(const int& s, const int& m);
	 /** Destructor. */
    virtual ~UnixSocket();
	 /** Assigment operator.
	  * Copies supplied socket fields into this object.
	  * @param s socket to copy.
	  */
    UnixSocket& operator=(const UnixSocket& s);
	 /** Access method.
	  * Allows setting of UnixSocket's internal socket descriptor. This is
	  * useful for allowing a UnixSocket object to encapsulate a socket
	  * descriptor that has been returned from another function, e.g.
	  * accept().
	  * @param s socket descriptor to encapsulate.
	  */
    void set_sockd(const int& s);
	 /** Socket API method.
	  * Binds socket to supplied address.
	  * @param path path to socket.
	  * @param port unused.
	  * @return 0: on success, -1 on failure.
	  */
    int bind(const string& path, const int& port=0);
	 /** Socket API method.
	  * Prepare a socket for accepting incoming connections.
	  * @return 0: on success, -1 on failure.
	  */
    int listen();
	 /** Socket API method.
	  * Wait for incoming connections on a socket. 
	  * @return -1: on failure, positive integer on success. On succes
	  *         the return value is a valid socket descriptor.
	  */
    int accept();
	 /** Socket API method.
	  * Attempt to establish a connection to the supplied address.
	  * @param path path to socket.
	  * @return 0: on success, -1 on failure.
	  */
    int connect(const string& path);
	 /** Socket API method.
	  * Attempt to receive data into supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes read.
	  */
    int recv(u_int8_t* b, const int& n);
	 /** Socket API method.
	  * Attempt to send data in supplied buffer.
	  * @param b socket buffer to store data in.
	  * @param n number of bytes to attempt to receive.
	  * @return 0: on success, positive integer on success. On success
	  *         the value returned is the number of bytes sent.
	  */
    int send(u_int8_t* b, const int& n);
    /** Socket API method.
	  * @doc shutdown connection as per standard socket library.
	  * @param how method to use to shutdown socket: SHUT_RD, SHUT_WR, 
	  * SHUT_RDWR.
	  */
    int shutdown(int howto);
	 /** Socket API method.
	  * @doc close socket connection.
	  */
    int close();
	 /** 
	  * @doc Test method.
	  */
    static bool test();
};

ostream& operator<<(ostream& os, const UnixSocket& s);

inline 
int UnixSocket::recv(u_int8_t* b, const int& n)
{
	 int length=0;
	 int iter=0;
	 int ret=0;
#if 0
	 fd_set rfd;
	 struct timeval timeout;
	 timeout.tv_sec=IO_TIMEOUT;
	 timeout.tv_usec=0;
	 int nd=_sockd+1;
	 do {
		  FD_SET(_sockd, &rfd);
		  ret=::select(nd, &rfd, NULL, NULL, &timeout);
		  if (ret==0) {
				syslog(LOG_ERR, "UnixSocket::recv(): timeout expired");
				return(-1);
		  }
		  if (ret<0) {
				syslog(LOG_ERR, "UnixSocket::recv(): select error: %s",
						 strerror(errno));
		  }
		  if (FD_ISSET(_sockd, &rfd)) {
				ret=::read(_sockd, b, n-length);
				if (ret>0) {
					 length+=ret;
					 if (length<n) {
						  b+=ret;
					 } else if (length>=n) {
						  return(length);
					 }
				} else if (ret<0 && errno!=EINTR) {
					 syslog(LOG_ERR, "UnixSocket::recv(): %s", strerror(errno));
					 return(-1);
				}
		  }
		  if (++iter>10) {
				syslog(LOG_ERR, "UnixSocket::recv(): exceeded max iter");
				return(length==0?-1:length);
		  }
	 } while ( (length<n) || (ret<0 && errno==EINTR) );
#else
	 do {
		  ret=read(_sockd, b, n-length);
		  ++_recv_calls;
		  if (ret>0) {
				_bytes_rcvd+=ret;
				length+=ret;
				if (length<n) {
					 b+=ret;
				} else if (length>=n) {
					 return(length);
				}
		  } else if (ret<0 && errno!=EINTR) {
				syslog(LOG_ERR, "UnixSocket::recv(): %s", strerror(errno));
				return(-1);
		  }
		  if (++iter>10) {
				syslog(LOG_ERR, "UnixSocket::recv(): exceeded max iter");
				return(length==0?-1:length);
		  }
	 } while ( (length<n) || (ret<0 && errno==EINTR) );
#endif

	 syslog(LOG_ERR, "UnixSocket::recv(): %s", strerror(errno));
	 return(-1);
}

inline 
int UnixSocket::send(u_int8_t* b, const int& n)
{
	 int sent=0;
	 int iter=0;
	 int ret=0;
#if 0
	 fd_set wfd;
	 struct timeval timeout;
	 timeout.tv_sec=IO_TIMEOUT;
	 timeout.tv_usec=0;
	 int nd=_sockd+1;
	 do {
		  FD_SET(_sockd, &wfd);
		  ret=::select(nd, NULL, &wfd, NULL, &timeout);
		  if (ret==0) {
				syslog(LOG_ERR, "UnixSocket::send(): select timeout");
				return(-1);
		  }
		  if (ret<0) {
				syslog(LOG_ERR, "UnixSocket::send(): select error: %s",
						 strerror(errno));
				return(-1);
		  }
		  if (FD_ISSET(_sockd, &wfd)) {
				ret=::write(_sockd, b, n-sent);
				if (ret>0) {
					 sent+=ret;
					 if (sent<n) {
						  b+=ret;
					 } else if (sent>=n) {
						  return(sent);
					 }
				} else if (ret<0 && errno!=EINTR) {
					 syslog(LOG_ERR, "UnixSocket::send(): %s", strerror(errno));
					 return(-1);
				}
		  }
		  if (++iter>10) {
				syslog(LOG_ERR, "UnixSocket::send(): exceeded max iter");
				return(sent==0?-1:sent);
		  }
	 } while ( (sent<n) || (ret<0 && errno==EINTR) );						 
#else
	 do {
		  ret=write(_sockd, b, n-sent);
		  ++_send_calls;
		  if (ret>0) {
				sent+=ret;
				_bytes_sent+=ret;
				if (sent<n) {
					 b+=ret;
				} else if (sent>=n) {
					 return(sent);
				}
		  } else if (ret<0 && errno!=EINTR) {
				syslog(LOG_ERR, "UnixSocket::send(): %s", strerror(errno));
				return(-1);
		  }
		  if (++iter>10) {
				syslog(LOG_ERR, "UnixSocket::send(): exceeded max iter");
				return(sent==0?-1:sent);
		  }
	 } while ( (sent<n) || (ret<0 && errno==EINTR) );						  
#endif
	 syslog(LOG_ERR, "UnixSocket::send(): %s", strerror(errno));
	 return(-1);
}


#endif

